# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/broeder-j/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################

"""
In this module contains useful methods for handling xml trees and files which are used
by the Fleur code and the fleur plugin.
"""
# TODO FEHLER meldungen, currently if a xpath expression is valid, but does not exists
# xpath returns []. Do we want this behavior?
# TODO finish implementation of create=False
# TODO: no aiida imports

from lxml import etree
from lxml.etree import XMLSyntaxError

from aiida.common.exceptions import InputValidationError

#from somewhere import ValidationError/InputValidationError
#some error, that does not depend on aiida

def is_sequence(arg):
    return (not hasattr(arg, "strip") and
            hasattr(arg, "__getitem__") or
            hasattr(arg, "__iter__"))

##### CONVERTERS ############
def convert_to_float(value_string, parser_info_out={'parser_warnings' : []}, suc_return=True):
    """
    Tries to make a float out of a string. If it can't it logs a warning
    and returns True or False if convertion worked or not.

    :param value_string: a string
    :returns value: the new float or value_string: the string given
    :retruns True or Falses
    """
    # TODO lowercase everything
    try:
        value = float(value_string)
    except TypeError:
        parser_info_out['parser_warnings'].append(
            'Could not convert: "{}" to float, TypeError'
            ''.format(value_string))
        if suc_return:
            return value_string, False
        else:
            return value_string
    except ValueError:
        parser_info_out['parser_warnings'].append(
            'Could not convert: "{}" to float, ValueError'
            ''.format(value_string))
        if suc_return:
            return value_string, False
        else:
            return value_string
    if suc_return:
        return value, True
    else:
        return value


def convert_to_int(value_string, parser_info_out={'parser_warnings' : []}, suc_return=True):
    """
    Tries to make a int out of a string. If it can't it logs a warning
    and returns True or False if convertion worked or not.

    :param value_string: a string
    :returns value: the new int or value_string: the string given
    :retruns True or False, if suc_return=True
    """
    try:
        value = int(value_string)
    except TypeError:
        parser_info_out['parser_warnings'].append(
            'Could not convert: "{}" to int, TypeError'
            ''.format(value_string))
        if suc_return:
            return value_string, False
        else:
            return value_string
    except ValueError:
        parser_info_out['parser_warnings'].append(
            'Could not convert: "{}" to int, ValueError'
            ''.format(value_string))
        if suc_return:
            return value_string, False
        else:
            return value_string
    if suc_return:
        return value, True
    else:
        return value



def convert_htr_to_ev(value, parser_info_out={'parser_warnings' : []}):
    """
    Multiplies the value given with the Hartree factor (converts htr to eV)
    """
    htr = 27.21138602
    suc = False
    value_to_save, suc = convert_to_float(value, parser_info_out=parser_info_out)
    if suc:
        return value_to_save*htr
    else:
        return value

def convert_ev_to_htr(value, parser_info_out={'parser_warnings' : []}):
    """
    Divides the value given with the Hartree factor (converts htr to eV)
    """
    htr = 27.21138602
    suc = False
    value_to_save, suc = convert_to_float(value, parser_info_out=parser_info_out)
    if suc:
        return value_to_save / htr
    else:
        return value


def convert_from_fortran_bool(stringbool):
    """
    Converts a string in this case ('T', 'F', or 't', 'f') to True or False

    :param stringbool: a string ('t', 'f', 'F', 'T')

    :return: boolean  (either True or False)
    """
    true_items = ['True', 't', 'T']
    false_items = ['False', 'f', 'F']
    if isinstance(stringbool, str):
        if stringbool in false_items:
            return False
        if stringbool in true_items:
            return True
        else:
            raise InputValidationError(
            "A string: {} for a boolean was given, which is not 'True',"
            " 'False', 't', 'T', 'F' or 'f'".format(stringbool))
    elif isinstance(stringbool, bool):
        return stringbool# no convertion needed...
    else:
        raise TypeError("convert_to_fortran_bool accepts only a string or "
                     "bool as argument")


def convert_to_fortran_bool(boolean):
    """
    Converts a Boolean as string to the format definded in the input

    :param boolean: either a boolean or a string ('True', 'False', 'F', 'T')

    :return: a string (either 't' or 'f')
    """

    if isinstance(boolean, bool):
        if boolean:
            new_string = 'T'
            return new_string
        else:
            new_string = 'F'
            return new_string
    elif isinstance(boolean, str):#basestring):
        if boolean == 'True' or boolean == 't' or boolean == 'T':
            new_string = 'T'
            return new_string
        elif boolean == 'False' or boolean == 'f' or boolean == 'F':
            new_string = 'F'
            return new_string
        else:
            raise InputValidationError(
                "A string: {} for a boolean was given, which is not 'True',"
                "'False', 't', 'T', 'F' or 'f'".format(boolean))
    else:
         raise TypeError("convert_to_fortran_bool accepts only a string or "
                         "bool as argument, given {} ".format(boolean))

def convert_to_fortran_string(string):
    """
    converts some parameter strings to the format for the inpgen
    :param string: some string
    :returns: string in right format (extra "")
    """
    new_string = '"' + string + '"'
    return new_string
    '''
    if isinstance(string, str):
        new_string = '"' + string + '"'
        return new_string
        #if '"' in string:
        #    return new_string
        #else:
        #    new_string = '"' + string + '"'
        #    return new_string
    else:
        #return string
        raise TypeError("_convert_to_fortran_string accepts only a"
              "string as argument, type {} given".format(type(string)))
    '''


def convert_fleur_lo(loelements):
    """
    Converts lo xml elements from the inp.xml file into a lo string for the inpgen
    """
    # Developer hint: Be careful with using '' and "", basestring and str are not the same...
    # therefore other conversion methods might fail, or the wrong format could be written.
    from aiida_fleur.tools.element_econfig_list import shell_map

    lo_string = ''
    for element in loelements:
        lo_type = get_xml_attribute(element, 'type')
        if lo_type != 'SCLO': # non standard los not supported for now
            continue
        l_num = get_xml_attribute(element, 'l')
        n_num = get_xml_attribute(element, 'n')
        l_char = shell_map.get(int(l_num), '')
        lostr = '{}{}'.format(n_num, l_char)
        lo_string = lo_string + ' ' + lostr
    return lo_string.strip()


def set_dict_or_not(para_dict, key, value):
    """
    setter method for a dictionary that will not set the key, value pair.
    if the key is [] or None.
    """
    if value==[] or value==None:
        return para_dict
    else:
        para_dict[key] = value
        return para_dict



####### XML SETTERS GENERAL ##############

'''
def xml_set_xpath(xmltree, xpathn, value):
    """
    """
    root = xmltree.getroot()
    nodes = root.xpath(xpathn)

    if type(attribv) != type(''):
        attribv = str(attribv)
    for node in nodes:
        node.set(attributename, attribv)
'''
'''
def set_xpath(xmltree, xpath, value, create=False):
    """
    method to set a general xpath in an xml file
    """
    # does not work
    # set text?

    # set attribute?
    nodes = eval_xpath3(xmltree, xpathn, create=create)
    for node in nodes:
        node = value
'''
def xml_set_attribv_occ(xmltree, xpathn, attributename, attribv, occ=[0], create=False):
    """
    Routine sets the value of an attribute in the xml file on only the places
    specified in occ

    :param: an etree a xpath from root to the attribute and the attribute value
    :param: occ, list of integers
    :return: None, or an etree

    Comment: Element.set will add the attribute if it does not exist,
             xpath expression has to exist
    example: xml_set_first_attribv(tree, '/fleurInput/calculationSetup', 'band', 'T')
             xml_set_first_attribv(tree, '/fleurInput/calculationSetup', 'dos', 'F')
    """

    root = xmltree.getroot()
    nodes = eval_xpath3(root, xpathn, create=create)
    #print('nodes from xml_set_attribv_occ: {}'.format(nodes))
    if type(attribv) != type(''):
        attribv = str(attribv)
    for i, node in enumerate(nodes):
        if i in occ:
            node.set(attributename, attribv)
        if -1 in occ:# 'all'
            node.set(attributename, attribv)

def xml_set_first_attribv(xmltree, xpathn, attributename, attribv, create=False):
    """
    Routine sets the value of an attribute in the xml file

    :param: an etree a xpath from root to the attribute and the attribute value

    :return: None, or an etree

    Comment: Element.set will add the attribute if it does not exist,
             xpath expression has to exist
    example: xml_set_first_attribv(tree, '/fleurInput/calculationSetup', 'band', 'T')
             xml_set_first_attribv(tree, '/fleurInput/calculationSetup', 'dos', 'F')
    """

    root = xmltree.getroot()
    if type(attribv) == type(''):
        eval_xpath3(root, xpathn, create=create)[0].set(attributename, attribv)
    else:
        eval_xpath3(root, xpathn, create=create)[0].set(attributename, str(attribv))
    #return xmltree
    #ToDO check if worked. else exception,

def xml_set_all_attribv(xmltree, xpathn, attributename, attribv, create=False):
    """
    Routine sets the value of an attribute in the xml file on all places it occurs

    :param: an etree a xpath from root to the attribute and the attribute value

    :return: None, or an etree

    Comment: Element.set will add the attribute if it does not exist,
             xpath expression has to exist
    example: xml_set_first_attribv(tree, '/fleurInput/atomGroups/atomGroup/force', 'relaxXYZ', 'TTF')
             xml_set_first_attribv(tree, '/fleurInput/atomGroups/atomGroup/force', 'calculate', 'F')
    """

    root = xmltree.getroot()
    nodes = eval_xpath3(root, xpathn, create=create)
    if is_sequence(attribv):
        for i, node in enumerate(nodes):
            if not isinstance(attribv[i], str):#type(attribv) != type(''):
                attribv[i] = str(attribv[i])
            node.set(attributename, attribv[i])
    else:
        if not isinstance(attribv, str):#type(attribv) != type(''):
            attribv = str(attribv)
        for node in nodes:
            node.set(attributename, attribv)

def xml_set_text(xmltree, xpathn, text, create=False, place_index=None, tag_order=None):
    """
    Routine sets the text of a tag in the xml file
    Input: an etree a xpath from root to the tag and the text
    Output: none, an etree
    example:
    xml_set_text(tree, '/fleurInput/comment', 'Test Fleur calculation for AiiDA plug-in')
    but also cordinates and Bravais Matrix!:
    xml_set_text(tree, '/fleurInput/atomGroups/atomGroup/relPos', '1.20000 PI/3 5.1-MYCrazyCostant')
    """

    root = xmltree.getroot()
    node = eval_xpath3(root, xpathn, create=create, place_index=place_index, tag_order=tag_order)
    if node:
        node[0].text = text
    #return xmltree

def xml_set_all_text(xmltree, xpathn, text, create=False):
    """
    Routine sets the text of a tag in the xml file
    Input: an etree a xpath from root to the tag and the text
    Output: none, an etree
    example:
    xml_set_text(tree, '/fleurInput/comment', 'Test Fleur calculation for AiiDA plug-in')
    but also cordinates and Bravais Matrix!:
    xml_set_text(tree, '/fleurInput/atomGroups/atomGroup/relPos', '1.20000 PI/3 5.1-MYCrazyCostant')
    """
    root = xmltree.getroot()
    nodes = eval_xpath3(root, xpathn, create=create)
    if is_sequence(text):
        for i, node in enumerate(nodes):
            node.text = text[i]
    else:
        for node in nodes:
            node.text = text

def create_tag(xmlnode, xpath, newelement, create=False, place_index = None, tag_order = None):
    """
    This method evaluates an xpath expresion and creates tag in an xmltree under the
    returned nodes. If the path does exist things will be overriden, or created.
    Per default the new element is appended to the elements, but it can also be
    inserted in a certain position or after certain other tags.

    """
    #root = xmltree.getroot()
    #print('create_tag {} {} {} {} {} {}'.format(xmlnode, xpath, newelement, create, place_index, tag_order))
    if not etree.iselement(newelement):
        #print('newelement from create_tag: {}'.format(newelement))
        #print('xpath from create_tag: {}'.format(xpath))
        try:
            newelement = etree.Element(newelement)
        except ValueError as v:
            raise ValueError('{}. If this is a species, are you sure this species exists in your inp.xml?'.format(v))
    nodes = eval_xpath3(xmlnode, xpath, create=create, place_index=place_index, tag_order=tag_order)
    #print('nodes found from create_tag: {}'.format(nodes))
    if nodes:
        for node_1 in nodes:
            if place_index:
                if tag_order:
                    # behind what shall I place it
                    behind_tags = tag_order[:place_index]
                    #children = node_1.getchildren()
                    # get all names of tag exisiting tags
                    set = False
                    for tag in reversed(behind_tags):
                        for child in node_1.iterchildren(tag=tag, reversed=False):
                            # if tagname of elements==tag:
                            tag_index = node_1.index(child)
                            try:
                                node_1.insert(tag_index, newelement)
                            except ValueError as v:
                                raise ValueError('{}. If this is a species, are'
                                    'you sure this species exists in your inp.xml?'
                                    ''.format(v))
                            set = True
                            break
                        if set:
                            break
                    if not set: # just append
                        try:
                            node_1.append(newelement)
                        except ValueError as v:
                            raise ValueError('{}. If this is a species, are you'
                                ' sure this species exists in your inp.xml?'
                                ''.format(v))
                    # (or remove all and write them again in right order?)
                else:
                    try:
                        node_1.insert(place_index, newelement)
                    except ValueError as v:
                        raise ValueError('{}. If this is a species, are you sure this species exists in your inp.xml?'.format(v))
            else:
                try:
                    node_1.append(newelement)
                except ValueError as v:
                    raise ValueError('{}. If this is a species, are you sure this species exists in your inp.xml?'.format(v))
    return xmlnode

def delete_att(xmltree, xpath, attrib):
    """
    deletes an xml tag in an xmletree in place

    param: xmltree: xmltree (etree)
    param: xpath: xpathexpression
    """
    root = xmltree.getroot()
    nodes = eval_xpath3(root, xpath)
    if nodes:
        for node in nodes:
            try:
                del node.attrib[attrib]
            except:
                pass
    return xmltree

def delete_tag(xmltree, xpath):
    """
    deletes an xml tag in an xmletree in place

    param: xmltree: xmltree (etree)
    param: xpath: xpathexpression
    """
    root = xmltree.getroot()
    nodes = eval_xpath3(root, xpath)
    if nodes:
        for node in nodes:
            parent = node.getparent()
            parent.remove(node)
    return xmltree

def replace_tag(xmltree, xpath, newelement):
    """
    replaces a xml tag by another tag on an xmletree in place

    param: xmltree: xmltree (etree)
    param: xpath: xpathexpression
    param: newelement: xmlElement
    """
    root = xmltree.getroot()

    nodes = eval_xpath3(root, xpath)
    if nodes:
        for node in nodes:
            parent = node.getparent()
            parent.remove(node)
            parent.append(newelement)

    return xmltree


def get_inpgen_paranode_from_xml(inpxmlfile):
    """
    This routine returns an AiiDA Parameter Data type produced from the inp.xml
    file, which can be used by inpgen.

    :return: ParameterData node
    """
    from aiida.orm.data.parameter import ParameterData
    para_dict = get_inpgen_para_from_xml(inpxmlfile)
    return ParameterData(dict=para_dict)

def get_inpgen_para_from_xml(inpxmlfile):
    """
    This routine returns an python dictionary produced from the inp.xml
    file, which can be used as a parameterdata node by inpgen.

    :return: dict

    Warning: This routine is prelimitary, it is far from complete.
    Also be aware that inpgen does not take all information that is contained in an inp.xml file
    """

    # TODO: convert econfig
    # TODO: parse kpoints, somehow count is bad (if symmetry changes), mesh is not known, path cannot be specified
    # TODO:
    #from aiida_fleur.tools.xml_util import eval_xpath, get_xml_attribute, convert_from_fortran_bool, eval_xpath2
    #from aiida_fleur.tools.xml_util import convert_to_float, convert_to_int

    #Disclaimer: this routine needs some xpath expressions. these are hardcoded here,
    #therefore maintainance might be needed, if you want to circumvent this, you have
    #to get all the paths from somewhere.

    #######
    # all hardcoded xpaths used and attributes names:
    # input
    film_xpath = '/fleurInput/atomGroups/atomGroup/filmPos/'# check for film pos

    # atom, for each species\
    species_xpath = '/fleurInput/atomSpecies/species'
    atom_id_xpath = '' # is reconstruction possible at all now?
    atom_z_xpath = '@atomicNumber'
    atom_rmt_xpath = 'mtSphere/@radius'
    atom_dx_xpath = 'mtSphere/@logIncrement'
    atom_jri_xpath = 'mtSphere/@gridPoints'
    atom_lmax_xpath = 'atomicCutoffs/@lmax'
    atom_lnosph_xpath = 'atomicCutoffs/@lnonsphr'
    atom_ncst_xpath = '@coreStates'
    atom_econfig_xpath = 'electronConfig' # converting todo
    atom_bmu_xpath = '@magMom'
    atom_lo_xpath = 'lo' # converting todo
    atom_element_xpath = '@element'
    atom_name_xpath = '@name'

    #comp
    jspins_xpath = 'calculationSetup/magnetism/@jspins'
    frcor_xpath = 'calculationSetup/coreElectrons/@frcor'
    ctail_xpath = 'calculationSetup/coreElectrons/@ctail'
    kcrel_xpath = 'calculationSetup/coreElectrons/@kcrel'
    gmax_xpath = 'calculationSetup/cutoffs/@Gmax'
    gmaxxc_xpath = 'calculationSetup/cutoffs/@GmaxXC'
    kmax_xpath = 'calculationSetup/cutoffs/@Kmax'

    #exco
    exco_xpath = 'xcFunctional/@name'
    # film

    #soc
    l_soc_xpath = '//calculationSetup/soc/@l_soc'
    theta_xpath = '//calculationSetup/soc/@theta'
    phi_xpath = '//calculationSetup/soc/@phi'
    # qss

    # kpt

    title_xpath = '/fleurInput/comment/text()' # text


    ########
    new_parameters = {}

    #print('parsing inp.xml without XMLSchema')
    tree = etree.parse(inpxmlfile)
    root = tree.getroot()

    # Create the cards

    # &input # most things are not needed for AiiDA here. or we ignor them for now.
    # film is set by the plugin depended on the structure
    #symor per default = False? to avoid input which fleur can't take

    # &comp
    #attrib = get_xml_attribute(
    comp_dict = {}
    comp_dict = set_dict_or_not(comp_dict, 'jspins', convert_to_float(eval_xpath(root, jspins_xpath), suc_return=False))
    comp_dict = set_dict_or_not(comp_dict, 'frcor', convert_from_fortran_bool(eval_xpath(root, frcor_xpath)))
    comp_dict = set_dict_or_not(comp_dict, 'ctail', convert_from_fortran_bool(eval_xpath(root, ctail_xpath)))#convert_from_fortran_bool()
    comp_dict = set_dict_or_not(comp_dict, 'kcrel', eval_xpath(root, kcrel_xpath))
    comp_dict = set_dict_or_not(comp_dict, 'gmax', convert_to_float(eval_xpath(root, gmax_xpath), suc_return=False))
    comp_dict = set_dict_or_not(comp_dict, 'gmaxxc', convert_to_float(eval_xpath(root, gmaxxc_xpath), suc_return=False))
    comp_dict = set_dict_or_not(comp_dict, 'kmax', convert_to_float(eval_xpath(root, kmax_xpath), suc_return=False))
    new_parameters['comp'] = comp_dict

    # &atoms
    species_list = eval_xpath2(root, species_xpath)

    for i, species in enumerate(species_list):
        atom_dict = {}
        atoms_name = 'atom{}'.format(i)
        atom_z = convert_to_int(eval_xpath(species, atom_z_xpath), suc_return=False)
        atom_rmt = convert_to_float(eval_xpath(species, atom_rmt_xpath), suc_return=False)
        atom_dx = convert_to_float(eval_xpath(species, atom_dx_xpath), suc_return=False)
        atom_jri = convert_to_int(eval_xpath(species, atom_jri_xpath), suc_return=False)
        atom_lmax = convert_to_int(eval_xpath(species, atom_lmax_xpath), suc_return=False)
        atom_lnosph = convert_to_int(eval_xpath(species, atom_lnosph_xpath), suc_return=False)
        atom_ncst = convert_to_int(eval_xpath(species, atom_ncst_xpath), suc_return=False)
        atom_econfig = eval_xpath(species, atom_econfig_xpath)
        atom_bmu = convert_to_float(eval_xpath(species, atom_bmu_xpath), suc_return=False)
        atom_lo = eval_xpath(species, atom_lo_xpath)
        atom_element = eval_xpath(species, atom_element_xpath)
        atom_name_2 = eval_xpath(species, atom_name_xpath)

        atom_dict = set_dict_or_not(atom_dict, 'z', atom_z)
        atom_dict = set_dict_or_not(atom_dict, 'rmt', atom_rmt)
        atom_dict = set_dict_or_not(atom_dict, 'dx', atom_dx)
        atom_dict = set_dict_or_not(atom_dict, 'jri', atom_jri)
        atom_dict = set_dict_or_not(atom_dict, 'lmax', atom_lmax)
        atom_dict = set_dict_or_not(atom_dict, 'lnonsph', atom_lnosph)
        atom_dict = set_dict_or_not(atom_dict, 'ncst', atom_ncst)
        atom_dict = set_dict_or_not(atom_dict, 'econfig', atom_econfig)
        atom_dict = set_dict_or_not(atom_dict, 'bmu', atom_bmu)
        if atom_lo is not None:
            atom_dict = set_dict_or_not(atom_dict, 'lo', convert_fleur_lo(atom_lo))
        atom_dict = set_dict_or_not(atom_dict, 'element', '{}'.format(atom_element))
        #atom_dict = set_dict_or_not(atom_dict, 'name', atom_name_2)

        new_parameters[atoms_name] = atom_dict


    # &soc
    attrib = convert_from_fortran_bool(eval_xpath(root, l_soc_xpath))
    theta = convert_to_float(eval_xpath(root, theta_xpath), suc_return=False)
    phi = convert_to_float(eval_xpath(root, phi_xpath), suc_return=False)
    if attrib:
        new_parameters['soc'] = {'theta' : theta, 'phi' : phi}

    # &kpt
    #attrib = convert_from_fortran_bool(eval_xpath(root, l_soc_xpath))
    #theta = eval_xpath(root, theta_xpath)
    #phi = eval_xpath(root, phi_xpath)
    #if kpt:
    #    new_parameters['kpt'] = {'theta' : theta, 'phi' : phi}
    #    # ['nkpt', 'kpts', 'div1', 'div2', 'div3',                         'tkb', 'tria'],

    # title
    title = eval_xpath(root, title_xpath)# text
    if title:
        new_parameters['title'] = title.replace('\n', '').strip()

    # &exco
    #TODO, easy
    exco_dict = {}
    exco_dict = set_dict_or_not(exco_dict, 'xctyp', eval_xpath(root, exco_xpath))
    #'exco' : ['xctyp', 'relxc'],
    new_parameters['exco'] = exco_dict
    # &film
    # TODO

    # &qss
    # TODO

    # lattice, not supported?

    return new_parameters

####### XML SETTERS SPECIAL ########

def set_species(fleurinp_tree_copy, species_name, attributedict, create=False):
    """
    Method to set parameters of a species tag of the fleur inp.xml file.

    param: fleurinp_tree_copy, xml etree of the inp.xml
    param: species_name : string, name of the species you want to change
    param: attributedict: python dict: what you want to change
    param: create: bool, if species does not exist create it and all subtags?

    raises: ValueError, if species name is non existent in inp.xml and should not be created.
    also if other given tags are garbage. (errors from eval_xpath() methods)

    return: fleurinp_tree_copy: xml etree of the new inp.xml


    """
    # TODO lowercase everything
    # TODO make a general specifier for species, not only the name i.e. also number, other parameters
    xpathspecies ='/fleurInput/atomSpecies/species[@name = "{}"]'.format(species_name)
    xpathmt = '{}/mtSphere'.format(xpathspecies)
    xpathatomicCutoffs = '{}/atomicCutoffs'.format(xpathspecies)
    xpathenergyParameters = '{}/energyParameters'.format(xpathspecies)
    xpathlo = '{}/lo'.format(xpathspecies)
    xpathelectronConfig = '{}/electronConfig'.format(xpathspecies)
    xpathcoreocc = '{}/electronConfig/stateOccupation'.format(xpathspecies)
    xpathLDA_U = '{}/ldaU'.format(xpathspecies)
    xpathnocoParams = '{}/nocoParams'.format(xpathspecies)
    xpathnocoParamsqss = '{}/nocoParams/qss'.format(xpathspecies)

    # can we get this out of schema file?
    species_seq = ['mtSphere', 'atomicCutoffs', 'energyParameters', 'force', 'electronConfig', 'nocoParams', 'ldaU', 'lo']

    #root = fleurinp_tree_copy.getroot()
    for key,val in attributedict.iteritems():
        if key == 'mtSphere': # always in inp.xml
            for attrib, value in val.iteritems():
                xml_set_attribv_occ(fleurinp_tree_copy, xpathmt, attrib, value)
        elif key == 'atomicCutoffs': # always in inp.xml
            for attrib, value in val.iteritems():
                xml_set_attribv_occ(fleurinp_tree_copy, xpathatomicCutoffs, attrib, value)
        elif key == 'energyParameters': # always in inp.xml
            for attrib, value in val.iteritems():
                xml_set_attribv_occ(fleurinp_tree_copy, xpathenergyParameters, attrib, value)
        elif key == 'lo': # optional in inp.xml
            #policy: we DELETE all LOs, and create new ones from the given parameters.
            existinglos = eval_xpath3(fleurinp_tree_copy, xpathlo)
            for los in existinglos:
                parent = los.getparent()
                parent.remove(los)

            # there can be multible LO tags, so I expect either one or a list
            if isinstance(val,dict):
                for attrib, value in val.iteritems():
                    xml_set_attribv_occ(fleurinp_tree_copy, xpathlo, attrib, value, create=create)
            else:# I expect a list of dicts
                #lonodes = eval_xpath3(root, xpathlo)#, create=True, place_index=species_seq.index('lo'), tag_order=species_seq)
                #nlonodes = len(lonodes)
                #ggf create more lo tags of needed
                los_need = len(val)# - nlonodes
                for j in range(0,los_need):
                    create_tag(fleurinp_tree_copy, xpathspecies, 'lo')#, place_index=species_seq.index('lo'), tag_order=species_seq)
                for i, lodict in enumerate(val):
                    for attrib, value in lodict.iteritems():
                        xml_set_attribv_occ(fleurinp_tree_copy, xpathlo, attrib, value, occ=[i])

        elif key == 'electronConfig':
            # eval electronConfig and ggf create tag at right place.
            #print('index {}'.format(species_seq.index('electronConfig')))
            eval_xpath3(fleurinp_tree_copy, xpathelectronConfig, create=True, place_index=species_seq.index('electronConfig'), tag_order=species_seq)

            for tag in ['coreConfig', 'valenceConfig', 'stateOccupation']:
                for etag, edictlist in val.iteritems():
                    if not etag == tag:
                        continue
                    if etag=='stateOccupation':# there can be multiple times stateOccupation
                        #policy: default we DELETE all existing occs and create new ones for the given input!
                        existingocc = eval_xpath3(fleurinp_tree_copy, xpathcoreocc)
                        for occ in existingocc:
                            parent = occ.getparent()
                            parent.remove(occ)
                        if isinstance(edictlist,dict):
                            #print('here')
                            for attrib, value in edictlist.iteritems():
                                xml_set_attribv_occ(fleurinp_tree_copy, xpathcoreocc, attrib, value, create=create)
                        else:# I expect a list of dicts
                            #occnodes = eval_xpath3(root, xpathcoreocc)
                            #noccnodes = len(occnodes)
                            #ggf create more lo tags of needed
                            nodes_need = len(edictlist)# - noccnodes
                            for j in range(0,nodes_need):
                                create_tag(fleurinp_tree_copy, xpathelectronConfig, 'stateOccupation', create=create)
                            for i, occdict in enumerate(edictlist):
                                #override them one after one
                                for attrib, value in occdict.iteritems():
                                    xml_set_attribv_occ(fleurinp_tree_copy, xpathcoreocc, attrib, value, occ=[i])

                    else:
                        xpathconfig = xpathelectronConfig + '/{}'.format(etag)
                        xml_set_text(fleurinp_tree_copy, xpathconfig, edictlist, create=create, place_index=species_seq.index('electronConfig'), tag_order = species_seq)
        elif key == 'nocoParams':
            for attrib, value in val.iteritems():
                if attrib == 'qss':
                    xml_set_text(fleurinp_tree_copy, xpathnocoParamsqss, attrib, value)
                else:
                   xml_set_attribv_occ(fleurinp_tree_copy, xpathnocoParams, attrib, value)
        elif key == 'ldaU':
            for attrib, value in val.iteritems():
                xml_set_attribv_occ(fleurinp_tree_copy, xpathLDA_U, attrib, value, create=create)
        else:
            xml_set_all_attribv(fleurinp_tree_copy, xpathspecies, attrib, value)

    return fleurinp_tree_copy

def add_lo(fleurinp_tree_copy, species_name, attributedict):
    pass


def change_atomgr_att(fleurinp_tree_copy, attributedict, position=None, species=None,create=False):

    xpathatmgroup = '/fleurInput/atomGroups/atomGroup'
    xpathforce = '{}/force'.format(xpathatmgroup)
    xpathnocoParams = '{}/nocoParams'.format(xpathatmgroup)

    if not position and not species: # not specfied what to change
        return fleurinp_tree_copy

    if position:
        if not position == 'all':
            xpathatmgroup ='/fleurInput/atomGroups/atomGroup/[/*Pos = {}]'.format(position)
            xpathforce = '{}/force'.format(xpathatmgroup)
            xpathnocoParams = '{}/nocoParams'.format(xpathatmgroup)
    if species:
        if not species == 'all':
            xpathatmgroup ='/fleurInput/atomGroups/atomGroup[@species = {}]'.format(species)
            xpathforce = '{}/force'.format(xpathatmgroup)
            xpathnocoParams = '{}/nocoParams'.format(xpathatmgroup)


    for key, val in attributedict.iteritems():
        if key == 'force':
            for attrib, value in val:
                xml_set_attribv_occ(fleurinp_tree_copy, xpathforce, attrib, value)
        elif key == 'nocoParams':
            for attrib, value in val:
                xml_set_attribv_occ(fleurinp_tree_copy, xpathnocoParams, attrib, value)
        else:
            xml_set_all_attribv(fleurinp_tree_copy, xpathatmgroup, attrib, value)



    return fleurinp_tree_copy

def add_num_to_att(xmltree, xpathn, attributename, set_val, mode='abs', occ=[0], create=False):
    """
    Routine adds something to the value of an attribute in the xml file (should be a number here)

    :param: an etree a xpath from root to the attribute and the attribute value

    :param: mode: 'abs', 'rel', change by absolut or relative amount
    :return: None, or an etree

    Comment: Element.set will add the attribute if it does not exist,
             xpath expression has to exist
    example: add_num_to_add(tree, '/fleurInput/bzIntegration', 'valenceElectrons', '1')
             add_num_to_add(tree, '/fleurInput/bzIntegration', 'valenceElectrons', '1.1', mode='rel')
    """


    #get attribute, add or multiply
    #set attribute
    attribval_node = eval_xpath(xmltree, xpathn)
    # do some checks..
    attribval = get_xml_attribute(attribval_node, attributename)
    print(attribval)
    if attribval:
        if mode=='abs':
            newattribv = float(attribval) + float(set_val)
        elif mode == 'rel':
            newattribv = float(attribval) * float(set_val)
        else:
            pass
            #unknown mode

        xml_set_attribv_occ(xmltree, xpathn, attributename, newattribv, occ=[0], create=False)
    else:
        pass
        # something was wrong, ...
    return xmltree

####### XML GETTERS #########
# TODO parser infos do not really work, might need to be returned, here
def eval_xpath(node, xpath, parser_info={'parser_warnings':[]}):
    """
    Tries to evalutate an xpath expression. If it fails it logs it.

    :param root node of an etree and an xpath expression (relative, or absolute)
    :returns either nodes, or attributes, or text
    """
    try:
        return_value = node.xpath(xpath)
    except etree.XPathEvalError:
        parser_info['parser_warnings'].append('There was a XpathEvalError on the xpath: {} \n'
            'Either it does not exist, or something is wrong with the expression.'.format(xpath))
        # TODO maybe raise an error again to catch in upper routine, to know where exactly
        return []
    if len(return_value) == 1:
        return return_value[0]
    else:
        return return_value


def eval_xpath2(node, xpath, parser_info={'parser_warnings':[]}):
    """
    Tries to evalutate an xpath expression. If it fails it logs it.

    :param root node of an etree and an xpath expression (relative, or absolute)
    :returns either nodes, or attributes, or text
    """
    try:
        return_value = node.xpath(xpath)
    except etree.XPathEvalError:
        parser_info['parser_warnings'].append('There was a XpathEvalError on the xpath: {} \n'
            'Either it does not exist, or something is wrong with the expression.'.format(xpath))
        # TODO maybe raise an error again to catch in upper routine, to know where exactly
        return []
    return return_value


'''
def eval_xpath2(node, xpath):
    """
    Tries to evalutate an xpath expression. If it fails it logs it.

    :param root node of an etree and an xpath expression (relative, or absolute)
    :returns ALWAYS a node list
    """
    try:
        return_value = node.xpath(xpath)
    except etree.XPathEvalError:
        print(
            'There was a XpathEvalError on the xpath: {} \n Either it does '
            'not exist, or something is wrong with the expression.'
            ''.format(xpath))
        return []
    return return_value
'''
def eval_xpath3(node, xpath, create=False, place_index=None, tag_order=None):
    """
    Tries to evalutate an xpath expression. If it fails it logs it.

    :param root node of an etree and an xpath expression (relative, or absolute)
    :returns ALWAYS a node list
    """
    #print('call eval_xpath3')
    try:
        return_value = node.xpath(xpath)
    except etree.XPathEvalError:
        message = (
            'There was a XpathEvalError on the xpath: {} \n Either it does '
            'not exist, or something is wrong with the expression.'
            ''.format(xpath))
        raise etree.XPathEvalError(message)
        return []

    if return_value == []:
        if create:
            x_pieces =  [e for e in xpath.split('/') if e != ""]
            #x_pieces = xpath.split('/')
            xpathn = ''
            #for piece in x_pieces[:-1]:
            #    piece = piece + '/'
            #    xpathn = xpathn + piece
            for piece in x_pieces[:-1]:
                xpathn = xpathn + '/'+ piece
            #node_c = eval_xpath3(node, xpathn, create)
            # this is REKURSIV! since create tag calls eval_xpath3
            create_tag(node, xpathn, x_pieces[-1], create=create, place_index=place_index, tag_order=tag_order)
            return_value = node.xpath(xpath)
            return return_value
        else:
            return return_value
    else:
        return return_value


def get_xml_attribute(node, attributename, parser_info_out={'parser_warnings' : []}):
    """
    Get an attribute value from a node.

    :param node: a node from etree
    :param attributename: a string with the attribute name.
    :returns either attributevalue, or None
    """
    if etree.iselement(node):
        attrib_value = node.get(attributename)
        if attrib_value:
            return attrib_value
        else:
            if parser_info_out:
                parser_info_out['parser_warnings'].append(
                    'Tried to get attribute: "{}" from element {}.\n '
                    'I recieved "{}", maybe the attribute does not exist'
                    ''.format(attributename, node, attrib_value))
            else:
                print(
                    'Can not get attributename: "{}" from node "{}", '
                    'because node is not an element of etree.'
                    ''.format(attributename, node))
            return None
    else: # something doesn't work here, some nodes get through here
        if parser_info_out:
            parser_info_out['parser_warnings'].append(
                'Can not get attributename: "{}" from node "{}", '
                'because node is not an element of etree.'
                ''.format(attributename, node))
        else:
            print(
                'Can not get attributename: "{}" from node "{}", '
                'because node is not an element of etree.'
                ''.format(attributename, node))
        return None




# TODO this has to be done better. be able to write tags and
# certain attributes of attributes that occur possible more then once.
# HINT: This is not really used anymore. use fleurinpmodifier
def write_new_fleur_xmlinp_file(inp_file_xmltree, fleur_change_dic, xmlinpstructure):
    """
    This modifies the xml-inp file. Makes all the changes wanted by
    the user or sets some default values for certain modes

    :param inp_file_lines_o xml-tree of the xml-inp file
    :param fleur_change_dic dictionary {attrib_name : value} with all the
           wanted changes.

    return an etree of the xml-inp file with changes.
    """
    # TODO rename, name is misleaded just changes the tree.
    xmltree_new = inp_file_xmltree

    #get all attributes of a inpxml file
    #xmlinpstructure = get_inpxml_file_structure()

    pos_switch_once = xmlinpstructure[0]
    pos_switch_several = xmlinpstructure[1]
    pos_attrib_once = xmlinpstructure[2]
    #pos_int_attributes_once = xmlinpstructure[3]
    pos_float_attributes_once = xmlinpstructure[4]
    #pos_string_attributes_once = xmlinpstructure[5]
    pos_attrib_several = xmlinpstructure[6]
    pos_int_attributes_several = xmlinpstructure[7]
    #pos_float_attributes_several = xmlinpstructure[8]
    #pos_string_attributes_several = xmlinpstructure[9]
    #pos_tags_several = xmlinpstructure[10]
    pos_text = xmlinpstructure[11]
    pos_xpaths = xmlinpstructure[12]
    expertkey = xmlinpstructure[13]


    for key in fleur_change_dic:
        if key in pos_switch_once:
            # call routine set (key,value) in tree.
            # TODO: a test here if path is plausible and if exist
            # ggf. create tags and key.value is 'T' or 'F' if not convert,
            # if garbage, exception
            # convert user input into 'fleurbool'
            fleur_bool = convert_to_fortran_bool(fleur_change_dic[key])

            xpath_set = pos_xpaths[key]
            #TODO: check if something in setup is inconsitent?

            # apply change to tree
            xml_set_first_attribv(xmltree_new, xpath_set, key, fleur_bool)

        elif key in pos_attrib_once:
            # TODO: same here, check existance and plausiblility of xpath
            xpath_set = pos_xpaths[key]
            if key in pos_float_attributes_once:
                newfloat = '{:.10f}'.format(fleur_change_dic[key])
                xml_set_first_attribv(xmltree_new, xpath_set, key, newfloat)
            else:
                xml_set_first_attribv(xmltree_new, xpath_set, key, fleur_change_dic[key])
        elif key in pos_attrib_several:
            # TODO What attribute shall be set? all, one or several specific onces?
            pass
        elif key in pos_switch_several:
            # TODO
            pass
        elif key in pos_text:
            # can be several times, therefore check
            xpath_set = pos_xpaths[key]
            xml_set_text(xmltree_new, xpath_set, fleur_change_dic[key])
        elif key == expertkey:
            # posibility for experts to set something, not suported by the plug-in directly
            pass
        else:
            # this key is not know to plug-in
            raise InputValidationError(
                "You try to set the key:'{}' to : '{}', but the key is unknown"
                " to the fleur plug-in".format(key, fleur_change_dic[key]))
    return xmltree_new


# TODO: maybe it is possible to use the xml, schema to dict libary of the QE people.
# So far it does not seem to do what we need.
def inpxml_todict(parent, xmlstr):
    """
    Recursive operation which transforms an xml etree to
    python nested dictionaries and lists.
    Decision to add a list is if the tag name is in the given list tag_several

    :param parent: some xmltree, or xml element
    :param xmlstr: structure/layout of the xml file in xmlstr is tags_several:
                   a list of the tags, which should be converted to a list, not
                   a dictionary(because they are known to occur more often, and
                   want to be accessed in a list later.

    :return: a python dictionary
    """

    xmlstructure = xmlstr
    pos_switch_once1 = xmlstructure[0]
    pos_switch_several1 = xmlstructure[1]
    #pos_attrib_once1 = xmlstructure[2]
    int_attributes_once1 = xmlstructure[3]
    float_attributes_once1 = xmlstructure[4]
    string_attributes_once1 = xmlstructure[5]
    #pos_attrib_several1 = xmlstructure[6]
    int_attributes_several1 = xmlstructure[7]
    float_attributes_several1 = xmlstructure[8]
    string_attributes_several1 = xmlstructure[9]
    tags_several1 = xmlstructure[10]
    pos_text1 = xmlstructure[11]
    #pos_xpaths1 = xmlstructure[12]
    #expertkey1 = xmlstructure[13]

    return_dict = {}
    if parent.items():
        return_dict = dict(parent.items())
        # Now we have to convert lazy fortan style into pretty things for the Database
        for key in return_dict:
            if key in pos_switch_once1 or (key in pos_switch_several1):
                return_dict[key] = convert_from_fortran_bool(return_dict[key])

            elif key in int_attributes_once1 or (key in int_attributes_several1):
                # TODO int several
                return_dict[key] = int(return_dict[key])
            elif key in string_attributes_once1 or (key in string_attributes_several1):
                # TODO What attribute shall be set? all, one or several specific onces?
                return_dict[key] = str(return_dict[key])
            elif key in float_attributes_once1 or (key in float_attributes_several1):
                # TODO pressision?
                return_dict[key] = float(return_dict[key])
                #pass
            elif key in pos_text1:
                # TODO, prob not nessesary, since taken care of below check,
                pass
            else:
                pass
                # this key is not know to plug-in TODO maybe make this a method
                # of the parser and log this as warning, or add here make a log
                # list, to which you always append messages, pass them back to
                # the parser, who locks it then
                # raise TypeError("Parser wanted to convert the key:'{}' with
                # value '{}', from the inpxml file but the key is unknown to the
                # fleur plug-in".format(key, return_dict[key]))


    if parent.text: # TODO more detal, exp: relPos
        # has text, but we don't want all the '\n' s and empty stings in the database
        if parent.text.strip() != '': # might not be the best solution
            # set text
            return_dict = parent.text.strip()

    firstocc = True
    for element in parent:
        if element.tag in tags_several1:
            #make a list, otherwise the tag will be overwritten in the dict
            if firstocc: # is this the first occurence?
                #create a list
                return_dict[element.tag] = []
                return_dict[element.tag].append(inpxml_todict(element, xmlstructure))
                firstocc = False
            else: # occured before, a list already exists, therefore just add
                return_dict[element.tag].append(inpxml_todict(element, xmlstructure))
        else:
            #make dict
            return_dict[element.tag] = inpxml_todict(element, xmlstructure)

    return return_dict


# This is probably only used to represent the whole inp.xml in the database for the fleurinpData attributes
# TODO this should be replaced by something else, maybe a class. that has a method to return certain
# list of possible xpaths from a schema file, or to validate a certain xpath expression and
# to allow to get SINGLE xpaths for certain attrbiutes.
#  akk: tell me where 'DOS' is
# This might not be back compatible... i.e a certain plugin version will by this design only work with
#  certain schema version
def get_inpxml_file_structure():
    """
    This routine returns the structure/layout of the 'inp.xml' file.

    Basicly the plug-in should know from this routine, what things are allowed
    to be set and where, i.e all attributes and their xpaths.
    As a developer make sure to use this routine always of you need information
    about the inp.xml file structure.
    Therefore, this plug-in should be easy to adjust to other codes with xml
    files as input files. Just rewrite this routine.

    For now the structure of the xmlinp file for fleur is hardcoded.
    If big changes are in the 'inp.xml' file, maintain this routine.
    TODO: Maybe this is better done, by reading the xml schema datei instead.
    And maybe it should also work without the schema file, do we want this?

    :param Nothing: TODO xml schema

    :return all_switches_once: list of all switches ('T' or 'F') which are allowed to be set
    :return all_switches_several: list of all switches ('T' or 'F') which are allowed to be set
    :return other_attributes_once: list of all attributes, which occur just once (can be tested)
    :return other_attributes_several: list of all attributes, which can occur more then once
    :return all_text: list of all text of tags, which can be set
    :return all_attrib_xpath: dictonary (attrib, xpath), of all possible attributes
                              with their xpath expression for the xmp inp
    :return expertkey: keyname (should not be in any other list), which can be
                       used to set anything in the file, by hand,
     (for experts, and that plug-in does not need to be directly maintained if
     xmlinp gets a new switch)
    """

    # All attributes (allowed to change?)

    #switches can be 'T' ot 'F' # TODO: alphabetical sorting
    all_switches_once = (
        'dos', 'band', 'secvar', 'ctail', 'frcor', 'l_noco',
        'ctail', 'swsp', 'lflip', 'off', 'spav', 'l_soc', 'soc66', 'pot8',
        'eig66', 'gamma', 'gauss', 'tria', 'invs', 'invs2', 'zrfs', 'vchk', 'cdinf',
        'disp', 'vacdos', 'integ', 'star', 'iplot', 'score', 'plplot', 'slice',
        'pallst', 'form66', 'eonly', 'bmt', 'relativisticCorrections', 'l_J', 'l_f')

    all_switches_several = ('calculate', 'flipSpin')

    int_attributes_once = ('numbands', 'itmax', 'maxIterBroyd', 'kcrel', 'jspins',
                           'gw', 'isec1', 'nx', 'ny', 'nz', 'ndir', 'layers',
                           'nstars', 'nstm', 'numkpt', 'nnne', 'lpr', 'count')
    float_attributes_once = ('Kmax', 'Gmax', 'GmaxXC', 'alpha', 'spinf', 'minDistance', 'theta', 
                             'phi', 'xa', 'thetad', 'epsdisp', 'epsforce', 'preconditioning_param',
                             'valenceElectrons', 'fermiSmearingEnergy', 'ellow',
                             'elup', 'scale', 'dTilda', 'dVac', 'minEnergy',
                             'maxEnergy', 'sigma', 'locx1', 'locy1', 'locx2',
                             'locy2', 'tworkf', 'minEigenval', 'maxEigenval')
    string_attributes_once = ('imix', 'mode', 'filename', 'latnam', 'spgrp',
                              'xcFunctional', 'fleurInputVersion', 'species')



    other_attributes_once = tuple(list(int_attributes_once) + list(float_attributes_once) + list(string_attributes_once))
    other_attributes_once1 = (
        'isec1', 'Kmax', 'Gmax', 'GmaxXC', 'numbands', 'itmax', 'maxIterBroyd',
        'imix', 'alpha', 'spinf', 'minDistance',
 'kcrel', 'jspins', 'theta', 'phi', 'gw', 'lpr',
        'xa', 'thetad', 'epsdisp', 'epsforce', 'valenceElectrons', 'mode',
        'gauss', 'fermiSmearingEnergy', 'nx', 'ny', 'nz', 'ellow', 'elup',
        'filename', 'scale', 'dTilda', 'dVac', 'ndir', 'minEnergy', 'maxEnergy',
        'sigma', 'layers', 'nstars', 'locx1', 'locy1', 'locx2', 'locy2', 'nstm',
        'tworkf', 'numkpt', 'minEigenval', 'maxEigenval', 'nnne')


    int_attributes_several = ('atomicNumber', 'gridPoints', 'lmax', 'lnonsphr',
                              's', 'p', 'd', 'f', 'l', 'n', 'eDeriv', 'coreStates')
    float_attributes_several = ('value', 'magMom', 'radius', 'logIncrement')
    string_attributes_several = ('name', 'element', 'coreStates', 'type', 'relaxXYZ')
    other_attributes_several = (
        'name', 'value', 'element', 'atomicNumber', 'coreStates', 'magMom',
        'radius', 'gridPoints', 'logIncrement', 'lmax', 'lnonsphr', 's', 'p',
        'd', 'f', 'species', 'type', 'coreStates', 'l', 'n', 'eDeriv', 'relaxXYZ')

    # when parsing the xml file to a dict, these tags should become
    # list(sets, or tuples) instead of dictionaries.
    tags_several = ('atomGroup', 'relPos', 'absPos', 'filmPos', 'species', 'kPoint')

    all_text = {'comment' : 1, 'relPos' : 3, 'filmPos' : 3, 'absPos' : 3,
                'row-1': 3, 'row-2':3, 'row-3' :3, 'a1': 1}
    #TODO all these (without comment) are floats, or float tuples.
    #Should be converted to this in the databas
    # changing the Bravais matrix should rather not be allowed I guess

    # all attribute xpaths

    #text xpaths(coordinates, bravaisMatrix)
    #all switches once, several, all attributes once, several
    all_attrib_xpath = {# text
        'comment': '/fleurInput/comment',
        'relPos': '/fleurInput/atomGroups/atomGroup/relPos',
        'filmPos': '/fleurInput/atomGroups/atomGroup/filmPos',
        'absPos': '/fleurInput/atomGroups/atomGroup/absPos',
        'row-1': '/fleurInput/cell/bulkLattice/bravaisMatrix',
        'row-2': '/fleurInput/cell/bulkLattice/bravaisMatrix',
        'row-3': '/fleurInput/cell/bulkLattice/bravaisMatrix',
        'a1': '/fleurInput/cell/filmLattice/a1', #switches once
        'dos': '/fleurInput/output',
        'band': '/fleurInput/output',
        'secvar': '/fleurInput/calculationSetup/expertModes',
        'ctail': '/fleurInput/calculationSetup/coreElectrons',
        'frcor': '/fleurInput/calculationSetup/coreElectrons',
        'l_noco': '/fleurInput/calculationSetup/magnetism',
        'l_J': '/fleurInput/calculationSetup/magnetism',
        'swsp': '/fleurInput/calculationSetup/magnetism',
        'lflip': '/fleurInput/calculationSetup/magnetism',
        'off': '/fleurInput/calculationSetup/soc',
        'spav': '/fleurInput/calculationSetup/soc',
        'l_soc': '/fleurInput/calculationSetup/soc',
        'soc66': '/fleurInput/calculationSetup/soc',
        'pot8': '/fleurInput/calculationSetup/expertModes',
        'eig66': '/fleurInput/calculationSetup/expertModes',
        'l_f': '/fleurInput/calculationSetup/geometryOptimization',
        'gamma': '/fleurInput/calculationSetup/bzIntegration/kPointMesh',
        'gauss': '',
        'tria': '',
        'invs': '',
        'zrfs': '',
        'vchk': '/fleurInput/output/checks',
        'cdinf': '/fleurInput/output/checks',
        'disp': '/fleurInput/output/checks',
        'vacdos': '/fleurInput/output',
        'integ': '/fleurInput/output/vacuumDOS',
        'star': '/fleurInput/output/vacuumDOS',
        'iplot': '/fleurInput/output/plotting',
        'score': '/fleurInput/output/plotting',
        'plplot': '/fleurInput/output/plotting',
        'slice': '/fleurInput/output',
        'pallst': '/fleurInput/output/chargeDensitySlicing',
        'form66': '/fleurInput/output/specialOutput',
        'eonly': '/fleurInput/output/specialOutput',
        'bmt': '/fleurInput/output/specialOutput',
        'relativisticCorrections': '/fleurInput/xcFunctional', #ALL_Switches_several
        'calculate': '/fleurInput/atomGroups/atomGroup/force',
        'flipSpin' : '/fleurInput/atomSpecies/species', #other_attributes_once
        'Kmax' : '/fleurInput/calculationSetup/cutoffs',
        'Gmax' : '/fleurInput/calculationSetup/cutoffs',
        'GmaxXC': '/fleurInput/calculationSetup/cutoffs',
        'numbands': '/fleurInput/calculationSetup/cutoffs',
        'itmax' : '/fleurInput/calculationSetup/scfLoop',
        'minDistance' : '/fleurInput/calculationSetup/scfLoop',
        'maxIterBroyd': '/fleurInput/calculationSetup/scfLoop',
        'preconditioning_param' : '/fleurInput/calculationSetup/scfLoop',
        'imix': '/fleurInput/calculationSetup/scfLoop',
        'alpha': '/fleurInput/calculationSetup/scfLoop',
        'spinf': '/fleurInput/calculationSetup/scfLoop',
        'kcrel': '/fleurInput/calculationSetup/coreElectrons',
        'jspins': '/fleurInput/calculationSetup/magnetism',
        'theta': '/fleurInput/calculationSetup/soc',
        'phi' : '/fleurInput/calculationSetup/soc',
        'gw': '/fleurInput/calculationSetup/expertModes',
        'lpr': '/fleurInput/calculationSetup/expertModes',
        'isec1': '/fleurInput/calculationSetup/expertModes',
        'xa': '/fleurInput/calculationSetup/geometryOptimization',
        'thetad': '/fleurInput/calculationSetup/geometryOptimization',
        'epsdisp': '/fleurInput/calculationSetup/geometryOptimization',
        'epsforce': '/fleurInput/calculationSetup/geometryOptimization',
        'valenceElectrons':'/fleurInput/calculationSetup/bzIntegration',
        'mode': '/fleurInput/calculationSetup/bzIntegration',
        'fermiSmearingEnergy': '/fleurInput/calculationSetup/bzIntegration',
        'nx': '/fleurInput/calculationSetup/bzIntegration/kPointMesh',
        'ny': '/fleurInput/calculationSetup/bzIntegration/kPointMesh',
        'nz': '/fleurInput/calculationSetup/bzIntegration/kPointMesh',
        'count': '/fleurInput/calculationSetup/kPointCount',
        'ellow' : '/fleurInput/calculationSetup/energyParameterLimits',
        'elup': '/fleurInput/calculationSetup',
        'filename': '/fleurInput/cell/symmetryFile',
        'scale': '/fleurInput/cell/bulkLattice',
        'ndir': '/fleurInput/output/densityOfStates',
        'minEnergy': '/fleurInput/output/densityOfStates',
        'maxEnergy': '/fleurInput/output/densityOfStates',
        'sigma':' /fleurInput/output/densityOfStates',
        'layers': '/fleurInput/output/vacuumDOS',
        'nstars': '/fleurInput/output/vacuumDOS',
        'locx1': '/fleurInput/output/vacuumDOS',
        'locy1': '/fleurInput/output/vacuumDOS',
        'locx2': '/fleurInput/output/vacuumDOS',
        'locy2': '/fleurInput/output/vacuumDOS',
        'nstm': '/fleurInput/output/vacuumDOS',
        'tworkf': '/fleurInput/output/vacuumDOS',
        'numkpts': '/fleurInput/output/chargeDensitySlicing',
        'minEigenval': '/fleurInput/output/chargeDensitySlicing',
        'maxEigenval': '/fleurInput/output/chargeDensitySlicing',
        'nnne': '/fleurInput/output/chargeDensitySlicing',
        'dVac': '/fleurInput/cell/filmLattice',
        'dTilda': '/fleurInput/cell/filmLattice',
        'xcFunctional': '/fleurInput/xcFunctional/@name',#other_attributes_more
        'name': {'/fleurInput/constantDefinitions', '/fleurInput/xcFunctional',
                 '/fleurInput/atomSpecies/species'},
        'value': '/fleurInput/constantDefinitions',
        'element': '/fleurInput/atomSpecies/species',
        'atomicNumber': '/fleurInput/atomSpecies/species',
        'coreStates': '/fleurInput/atomSpecies/species',
        'magMom': '/fleurInput/atomSpecies/species',
        'radius': '/fleurInput/atomSpecies/species/mtSphere',
        'gridPoints': '/fleurInput/atomSpecies/species/mtSphere',
        'logIncrement': '/fleurInput/atomSpecies/species/mtSphere',
        'lmax': '/fleurInput/atomSpecies/species/atomicCutoffs',
        'lnonsphr': '/fleurInput/atomSpecies/species/atomicCutoffs',
        's': '/fleurInput/atomSpecies/species/energyParameters',
        'p': '/fleurInput/atomSpecies/species/energyParameters',
        'd': '/fleurInput/atomSpecies/species/energyParameters',
        'f': '/fleurInput/atomSpecies/species/energyParameters',
        'type': '/fleurInput/atomSpecies/species/lo',
        'l': '/fleurInput/atomSpecies/species/lo',
        'n': '/fleurInput/atomSpecies/species/lo',
        'eDeriv': '/fleurInput/atomSpecies/species/lo',
        'species': '/fleurInput/atomGroups/atomGroup',
        'relaxXYX': '/fleurInput/atomGroups/atomGroup/force'
    }
    #'constant': (name, value)
    #'xcFunctional' : name
    #'species' : (old_name, set_name)
    #'radius': (spcies_name, value), (species_id, value), (species_element, value)
    # all tags paths, make this a dict?

    all_tag_xpaths = (
        '/fleurInput/constantDefinitions',
        '/fleurInput/calculationSetup',
        '/fleurInput/calculationSetup/cutoffs',
        '/fleurInput/calculationSetup/scfLoop',
        '/fleurInput/calculationSetup/coreElectrons',
        '/fleurInput/calculationSetup/magnetism',
        '/fleurInput/calculationSetup/soc',
        '/fleurInput/calculationSetup/expertModes',
        '/fleurInput/calculationSetup/geometryOptimization',
        '/fleurInput/calculationSetup/bzIntegration',
        '/fleurInput/calculationSetup/kPointMesh',
        '/fleurInput/cell/symmetry',
        '/fleurInput/cell/bravaisMatrix',
        '/fleurInput/xcFunctional',
        '/fleurInput/xcFunctional/xcParams',
        '/fleurInput/atomSpecies/species',
        '/fleurInput/atomSpecies/species/mtSphere',
        '/fleurInput/atomSpecies/species/atomicCutoffs',
        '/fleurInput/atomSpecies/species/energyParameters',
        '/fleurInput/atomSpecies/species/coreConfig',
        '/fleurInput/atomSpecies/species/coreOccupation',
        '/fleurInput/atomGroups/atomGroup',
        '/fleurInput/atomGroups/atomGroup/relPos',
        '/fleurInput/atomGroups/atomGroup/absPos',
        '/fleurInput/atomGroups/atomGroup/filmPos',
        '/fleurInput/output/checks',
        '/fleurInput/output/densityOfStates',
        '/fleurInput/output/vacuumDOS',
        '/fleurInput/output/plotting',
        '/fleurInput/output/chargeDensitySlicing',
        '/fleurInput/output/specialOutput'
    )

    expertkey = 'other'
    returnlist = (all_switches_once,
                  all_switches_several,
                  other_attributes_once,
                  int_attributes_once,
                  float_attributes_once,
                  string_attributes_once,
                  other_attributes_several,
                  int_attributes_several,
                  float_attributes_several,
                  string_attributes_several,
                  tags_several,
                  all_text,
                  all_attrib_xpath,
                  expertkey)
    return returnlist





# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/broeder-j/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################

"""
In here we put all things util (methods, code snipets) that are often useful, but not yet in AiiDA
itself.
So far it contains:

export_extras
import_extras
delete_nodes
delete_trash
create_group

"""
# TODO import, export of descriptions, and labels...?
import json

from aiida.orm import Code, DataFactory, load_node
from aiida.orm.querybuilder import QueryBuilder, Node
from aiida.orm import Group


RemoteData = DataFactory('remote')
ParameterData = DataFactory('parameter')
FleurInpData = DataFactory('fleur.fleurinp')



def export_extras(nodes, filename='node_extras.txt'):
    """
    writes uuids and extras of given nodes to a file (json).
    This is useful for import/export because currently extras are lost.
    Therefore this can be used to save and restore the extras on the nodes.

    :param: nodes: list of AiiDA nodes, pks, or uuids
    :param: filename, string where to store the file and its name

    example use:
    node_list = [120,121,123,46]
    export_extras(node_list)
    """

    #outstring = ''#' node uuid | extras \n'
    outdict = {}
    for node in nodes:
        if isinstance(node, int): #pk
            node = load_node(node)
        elif isinstance(node, basestring): #uuid
            node = load_node(node)

        if not isinstance(node, Node):
            print('skiped node {}, is not an AiiDA node, did not know what to do.'.format(node))
            continue
        uuid = node.uuid
        extras_dict = node.get_extras()
        outdict[uuid] = extras_dict
        #line = '{} | {}\n'.format(uuid, extras_dict)
        #outstring = outstring + line

    #outfile = open(filename, 'w')
    #outfile.write(outstring)
    #outfile.close()
    json.dump(outdict, open(filename,'w'))
    return


def import_extras(filename):
    """
    reads in nodes uuids and extras from a file and aplies them to nodes in the DB.

    This is useful for import/export because currently extras are lost.
    Therefore this can be used to save and restore the extras on the nodes.

    :param: filename, string what file to read from (has to be json format)

    example use:
    import_extras('node_extras.txt')
    """

    all_extras = {}

    # read file
    #inputfile = open(filename, 'r')
    #lines = inputfile.readlines()
    #for line in lines[1:]:
    #    splitted = line.split(' | ')
    #    uuid = splitted[0].rstrip(' ')
    #    extras = splitted[1].rstrip(' ')
    #    #extras = dict(extras)
    #    print(extras)
    #    all_extras[uuid] = extras
    #inputfile.close()
    try:
        all_extras = json.load(open(filename))
    except:
        print('The file has to be loadabel by json. i.e json format (which it is not).')

    for uuid, extras in all_extras.iteritems():

        try:
            node = load_node(uuid)
        except:
            # Does not exists
            print('node with uuid {} does not exist in DB'.format(uuid))
            node = None
            continue
        if isinstance(node, Node):
            node.set_extras(extras)
        else:
            print('node is not instance of an AiiDA node')
        #print(extras)
    return



def delete_nodes(pks_to_delete):
    """
    Delete a set of nodes. (From AiiDA cockbook)
    Note: TODO this has to be improved for workfchain removal. (checkpoints and co)
    Also you will be backchecked.

    BE VERY CAREFUL!
    TODO: CHECK IF THIS IS UP TO DATE!
    
    :note: The script will also delete
    all children calculations generated from the specified nodes.

    :param pks_to_delete: a list of the PKs of the nodes to delete
    """
    from django.db import transaction
    from django.db.models import Q
    from aiida.backends.djsite.db import models
    from aiida.orm import load_node

    # Delete also all children of the given calculations
    # Here I get a set of all pks to actually delete, including
    # all children nodes.
    all_pks_to_delete = set(pks_to_delete)
    for pk in pks_to_delete:
        all_pks_to_delete.update(models.DbNode.objects.filter(
            parents__in=pks_to_delete).values_list('pk', flat=True))

    print("I am going to delete {} nodes, including ALL THE CHILDREN"
          "of the nodes you specified. Do you want to continue? [y/N]"
          "".format(len(all_pks_to_delete)))
    answer = raw_input()

    if answer.strip().lower() == 'y':
        # Recover the list of folders to delete before actually deleting
        # the nodes.  I will delete the folders only later, so that if
        # there is a problem during the deletion of the nodes in
        # the DB, I don't delete the folders
        folders = [load_node(pk).folder for pk in all_pks_to_delete]

        with transaction.atomic():
            # Delete all links pointing to or from a given node
            models.DbLink.objects.filter(
                Q(input__in=all_pks_to_delete) |
                Q(output__in=all_pks_to_delete)).delete()
            # now delete nodes
            models.DbNode.objects.filter(pk__in=all_pks_to_delete).delete()

        # If we are here, we managed to delete the entries from the DB.
        # I can now delete the folders
        for f in folders:
            f.erase()

def delete_trash():
    """
    This method deletes all AiiDA nodes in the DB, which have a extra trash=True
    And all their children. Could be advanced to a garbage collector.

    Be careful to use it.
    """

    #query db for marked trash
    q = QueryBuilder()
    nodes_to_delete_pks = []

    q.append(Node,
            filters = {'extras.trash': {'==' : True}
                       }
            )
    res = q.all()
    for node in res:
        nodes_to_delete_pks.append(node[0].dbnode.pk)
        print('pk {}, extras {}'.format(node[0].dbnode.pk, node[0].get_extras()))

    #Delete the trash nodes

    print('deleting nodes {}'.format(nodes_to_delete_pks))
    delete_nodes(nodes_to_delete_pks)

    return

def create_group(name, nodes, description=None):
    """
    Creates a group for a given node list.

    So far this is only an AiiDA verdi command.
    :param name: string name for the group
    :param nodes: list of AiiDA nodes, pks, or uuids
    :param description, optional string that will be stored as description for the group

    :return: the group, AiiDa group
    usage example:

    group_name = 'delta_structures_gustav'
    nodes_to_goup_pks =[2142, 2084]
    create_group(group_name, nodes_to_group_pks, description='delta structures added by hand. from Gustavs inpgen files')
    """
    group, created = Group.get_or_create(name=name)
    if created:
        print('Group created with PK={} and name {}'.format(group.pk, group.name))
    else:
        print('Group with name {} and pk {} already exists. Do you want to add nodes?[y/n]'.format(group.name, group.pk))
        answer = raw_input()
        if answer.strip().lower() == 'y':
            pass
        else:
            return
    nodes2 = []
    nodes2_pks = []
    for node in nodes:
        try:
            node = int(node)
        except ValueError:
            pass
        nodes2_pks.append(node)
        try:
            nodes2.append(load_node(node))
        except:# NotExistentError:
            pass

    group.add_nodes(nodes2)
    print('added nodes: {} to group {} {}'.format(nodes2_pks, group.name, group.pk))

    if description:
        group.description = description

    return group


def get_nodes_from_group(group, return_format='uuid'):
    """
    returns a list of node uuids for a given group as, name, pk, uuid or group object
    """
    from aiida.orm import Group
    from aiida.common.exceptions import NotExistent

    nodes = []
    g_nodes = []


    try:
        group_pk = int(group)
    except ValueError:
        group_pk = None
        group_name = group

    if group_pk is not None:
        try:
            str_group = Group(dbgroup=group_pk)
        except NotExistent:
            str_group = None
            message = ('You have to provide a valid pk for a Group '
                       'or a Group name. Reference key: "group".'
                      'given pk= {} is not a valid group'
                      '(or is your group name integer?)'.format(group_pk))
            print(message)
    elif group_name is not None:
        try:
            str_group = Group.get_from_string(group_name)
        except NotExistent:
            str_group = None
            message = ('You have to provide a valid pk for a Group or a Group name.'
                      'given group name= {} is not a valid group'
                      '(or is your group name integer?)'.format(group_name))
            print(message)
    elif isinstance(group, Group):
        str_group = group
    else:
        str_group = None
        print('I could not handle given input, either Group, pk, or group name please.')
        return nodes

    g_nodes = str_group.nodes

    for node in g_nodes:
        if return_format == 'uuid':
            nodes.append(node.uuid)
        elif return_format == 'pk':
            nodes.append(node.pk)
        elif return_format == 'node':
            nodes.append(node)

    return nodes






# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/broeder-j/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################

"""
In this module you find the workflow 'fleur_eos_wc' for the calculation of
of an equation of state
"""
#TODO: print more user info
#  allow different inputs, make things optional(don't know yet how)
#  half number of iteration if you are close to be converged. (therefore one can start with 18 iterations, and if thats not enough run agian 9 or something)
#from sys import argv
#import time
import numpy as np
from aiida.orm import Code, DataFactory, load_node
from aiida.orm.data.base import Float
from aiida.work.workchain import WorkChain, ToContext#,Outputs
from aiida.work import workfunction as wf
from aiida.work.run import submit
from aiida_fleur.tools.StructureData_util import rescale, is_structure
from aiida_fleur.workflows.scf import fleur_scf_wc
from aiida_fleur.tools.common_fleur_wf import test_and_get_codenode
from aiida_fleur.tools.common_fleur_wf_util import check_eos_energies

StructureData = DataFactory('structure')
ParameterData = DataFactory('parameter')
FleurInpData = DataFactory('fleur.fleurinp')


class fleur_eos_wc(WorkChain):
    """
    This workflow calculates the equation of states of a structure.
    Calculates several unit cells with different volumes.
    A Birch_Murnaghan  equation of states fit determines the Bulk modulus and the
    groundstate volume of the cell.

    :param wf_parameters: ParameterData node, optional 'wf_parameters', protocol specifieing parameter dict
    :param structure: StructureData node, 'structure' crystal structure
    :param calc_parameters: ParameterData node, optional 'calc_parameters' parameters for inpgen
    :param inpgen: Code node,
    :param fleur: Code node,


    :return output_eos_wc_para: ParameterData node, contains relevant output information.
    about general succces, fit results and so on.


    example input.
    """

    _workflowversion = "0.3.3"

    _default_options = {
                        u'resources' : {"num_machines": 1},
                        u'max_wallclock_seconds' : 6*60*60,
                        u'queue_name' : u'',
                        u'custom_scheduler_commands' : u'',
                        u'import_sys_environment' : False,
                        u'environment_variables' : {}}    

                        
    ERROR_INVALID_INPUT_RESOURCES = 1
    ERROR_INVALID_INPUT_RESOURCES_UNDERSPECIFIED = 2
    ERROR_INVALID_CODE_PROVIDED = 3
    ERROR_INPGEN_CALCULATION_FAILED = 4
    ERROR_CHANGING_FLEURINPUT_FAILED = 5
    ERROR_CALCULATION_INVALID_INPUT_FILE = 6
    ERROR_FLEUR_CALCULATION_FALIED = 7
    ERROR_CONVERGENCE_NOT_ARCHIVED = 8
    
    
    def __init__(self, *args, **kwargs):
        super(fleur_eos_wc, self).__init__(*args, **kwargs)


    @classmethod
    def define(cls, spec):
        super(fleur_eos_wc, cls).define(spec)
        spec.input("wf_parameters", valid_type=ParameterData, required=False,
                   default=ParameterData(dict={
                       'fleur_runmax': 4,
                       'points' : 9,
                       'step' : 0.002,
                       'guess' : 1.00}))
        spec.input("structure", valid_type=StructureData, required=True)
        spec.input("calc_parameters", valid_type=ParameterData, required=False)
        spec.input("inpgen", valid_type=Code, required=True)
        spec.input("fleur", valid_type=Code, required=True)
        spec.input("options", valid_type=ParameterData, required=False, 
                   default=ParameterData(dict={
                            'resources': {"num_machines": 1},
                            'max_wallclock_seconds': 60*60,
                            'queue_name': '',
                            'custom_scheduler_commands' : '',
                            'import_sys_environment' : False,
                            'environment_variables' : {}}))
        spec.input("settings", valid_type=ParameterData, required=False)

        spec.outline(
            cls.start,
            cls.structures,
            cls.converge_scf,
            cls.return_results
        )

        
    def start(self):
        """
        check parameters, what condictions? complete?
        check input nodes
        """
        self.report("Started eos workflow version {} "
                    #"Workchain node identifiers: {}"
                    "".format(self._workflowversion))#, self.uuid))

        
        ### input check ### ? or done automaticly, how optional?
        self.ctx.last_calc2 = None
        self.ctx.calcs = []
        self.ctx.calcs_future = []
        self.ctx.structures = []
        self.ctx.temp_calc = None
        self.ctx.structurs_uuids = []
        self.ctx.scalelist = []
        self.ctx.volume = []
        self.ctx.volume_peratom = []
        self.ctx.org_volume = -1# avoid div 0
        self.ctx.labels = []
        self.ctx.successful = True#False
        self.ctx.info = []
        self.ctx.warnings = []
        self.ctx.errors = []
        # TODO get all succesfull from convergence, if all True this


        wf_dict = self.inputs.wf_parameters.get_dict()
        # set values, or defaults, default: always converge charge density,
        # crit < 0.00002, max 4 fleur runs

        self.ctx.points = wf_dict.get('points', 9)
        self.ctx.step = wf_dict.get('step', 0.002)
        self.ctx.guess = wf_dict.get('guess', 1.00)
        self.ctx.serial = wf_dict.get('serial', False)#True
        self.ctx.max_number_runs = wf_dict.get('fleur_runmax', 4)

        #defaultoptions = self._default_options
        #options = wf_dict.get('options', defaultoptions)
        #if 'options' in self.inputs:
        #    options = self.inputs.options.get_dict()
        #else:
        #    options = defaultoptions
        #for key, val in defaultoptions.iteritems():
        #    options[key] = options.get(key, val)
        #self.ctx.options = options
        
        inputs = self.inputs

        if 'inpgen' in inputs:
            try:
                test_and_get_codenode(inputs.inpgen, 'fleur.inpgen', use_exceptions=True)
            except ValueError:
                error = ("The code you provided for inpgen of FLEUR does not "
                         "use the plugin fleur.inpgen")
                self.control_end_wc(error)
                return self.ERROR_INVALID_CODE_PROVIDED

        if 'fleur' in inputs:
            try:
                test_and_get_codenode(inputs.fleur, 'fleur.fleur', use_exceptions=True)
            except ValueError:
                error = ("The code you provided for FLEUR does not "
                         "use the plugin fleur.fleur")
                self.control_end_wc(error)
                return self.ERROR_INVALID_CODE_PROVIDED

    def structures(self):
        """
        Creates structure data nodes with different Volume (lattice constants)
        """
        points = self.ctx.points
        step = self.ctx.step
        guess = self.ctx.guess
        startscale = guess-(points-1)/2*step

        for point in range(points):
            self.ctx.scalelist.append(startscale + point*step)

        self.report('scaling factors which will be calculated:{}'.format(self.ctx.scalelist))
        self.ctx.org_volume = self.inputs.structure.get_cell_volume()
        self.ctx.structurs = eos_structures(self.inputs.structure, self.ctx.scalelist)

    def converge_scf(self):
        """
        Launch fleur_scfs from the generated structures
        """
        calcs = {}

        for i, struc in enumerate(self.ctx.structurs):
            inputs = self.get_inputs_scf()
            inputs['structure'] = struc
            natoms = len(struc.sites)
            label = str(self.ctx.scalelist[i])
            label_c = '|eos| fleur_scf_wc'
            description = '|fleur_eos_wc|fleur_scf_wc|scale {}, {}'.format(label, i)
            inputs['_label'] = label_c
            inputs['_description'] = description
            
            self.ctx.volume.append(struc.get_cell_volume())
            self.ctx.volume_peratom.append(struc.get_cell_volume()/natoms)
            self.ctx.structurs_uuids.append(struc.uuid)

            #calc_para = inputs['calc_parameters']
            
            # Here we give the same wf_parameters to the fleur_scf...
            #if calc_para:
            #     res = self.submit(fleur_scf_wc,
            #                  wf_parameters=inputs['wf_parameters'],
            #                  structure=inputs['structure'],
            #                  calc_parameters=calc_para,
            #                  inpgen=inputs['inpgen'],
            #                  fleur=inputs['fleur'],
            #                  label=label_c,
            #                  description=description)
            # else:
            #     res = self.submit(fleur_scf_wc,
            #                  wf_parameters=inputs['wf_parameters'],
            #                  structure=inputs['structure'],
            #                  inpgen=inputs['inpgen'],
            #                  fleur=inputs['fleur'],
            #                  label=label_c,
            #                  description=description)
            #time.sleep(5)
            res = submit(fleur_scf_wc, **inputs)
            self.ctx.labels.append(label)
            calcs[label] = res

        return ToContext(**calcs)


    def get_inputs_scf(self):
        """
        get and 'produce' the inputs for a scf-cycle
        """
        inputs = {}

        # create input from that
        wf_para_dict = self.inputs.wf_parameters.get_dict()
        if wf_para_dict.get('scf_para', None):
            inputs['wf_parameters'] = wf_para_dict.get('scf_para', None)

        #if not inputs['wf_parameters']:
        #    para = {}
        #    para['options'] = wf_para_dict.get('options')            
        #    #para['resources'] = wf_para_dict.get('resources')
        #    #para['walltime_sec'] = wf_para_dict.get('walltime_sec')
        #    #para['queue_name'] = wf_para_dict.get('queue_name')
        #    para['serial'] = wf_para_dict.get('serial')
        #    #para['custom_scheduler_commands'] = wf_para_dict.get('custom_scheduler_commands')
        #    inputs['wf_parameters'] = ParameterData(dict=para)
        if 'options' in self.inputs:
            inputs['options'] = self.inputs.options
        try:
            calc_para = self.inputs.calc_parameters
        except AttributeError:
            calc_para = None

        if calc_para:
            inputs['calc_parameters'] = calc_para
        inputs['inpgen'] = self.inputs.inpgen
        inputs['fleur'] = self.inputs.fleur

        return inputs


    def return_results(self):
        """
        return the results of the calculations  (scf workchains) and do a
        Birch-Murnaghan fit for the equation of states
        """
        distancelist = []
        t_energylist = []
        t_energylist_peratom = []
        outnodedict = {}
        natoms = len(self.inputs.structure.sites)
        htr2eV = 27.21138602

        for label in self.ctx.labels:
            calc = self.ctx[label]
            try:
                outnodedict[label] = calc.get_outputs_dict()['output_scf_wc_para']
            except KeyError:
                message = ('One SCF workflow failed, no scf output node: {}. I skip this one.'.format(label))
                self.ctx.errors.append(message)
                self.ctx.successful = False
                continue
            
            outpara = calc.get_outputs_dict()['output_scf_wc_para'].get_dict()

            if not outpara.get('successful', False):
                #TODO: maybe do something else here
                # (exclude point and write a warning or so, or error treatment)
                # bzw implement and scf_handler,
                #also if not perfect converged, results might be good
                message = ('One SCF workflow was not successful: {}'.format(label))
                self.ctx.warning.append(message)
                self.ctx.successful = False
                
            t_e = outpara.get('total_energy', float('nan'))
            e_u = outpara.get('total_energy_units', 'eV')
            if e_u == 'Htr' or 'htr':
                t_e = t_e * htr2eV
            dis = outpara.get('distance_charge', float('nan'))
            dis_u = outpara.get('distance_charge_units')
            t_energylist.append(t_e)
            t_energylist_peratom.append(t_e/natoms)
            distancelist.append(dis)
        
        not_ok, an_index = check_eos_energies(t_energylist_peratom)
        
        if not_ok:
            message = ('Abnormality in Total energy list detected. Check '
                       'entr(ies) {}.'.format(an_index))
            hint = ('Consider refining your basis set.')
            self.ctx.info.append(hint)
            self.ctx.warnings.append(message)
            
            
        a = np.array(t_energylist_peratom)
        b = np.array(self.ctx.volume_peratom)

        # all erros should be caught before\
        
        # TODO: different fits
        volume, bulk_modulus, bulk_deriv, residuals = birch_murnaghan_fit(a, b)
        
        volumes = self.ctx.volume
        gs_scale = volume*natoms/self.ctx.org_volume
        if (volume*natoms < volumes[0]) or (volume*natoms > volumes[-1]):
            warn = ('Groundstate volume was not in the scaling range.')         
            hint = ('Consider rerunnning around point {}'.format(gs_scale))
            self.ctx.info.append(hint)
            self.ctx.warnings.append(warn)
            # TODO maybe make it a feature to rerun with centered around the gs.
        
        out = {'workflow_name' : self.__class__.__name__,
               'workflow_version' : self._workflowversion,
               'scaling': self.ctx.scalelist,
               'scaling_gs' : gs_scale,
               'initial_structure': self.inputs.structure.uuid,
               'volume_gs' : volume*natoms,
               'volumes' : volumes,
               'volume_units' : 'A^3',
               'natoms' : natoms,
               'total_energy': t_energylist,
               'total_energy_units' : e_u,
               'structures' : self.ctx.structurs_uuids,
               'calculations' : [],#self.ctx.calcs1,
               'scf_wfs' : [],#self.converge_scf_uuids,
               'distance_charge' : distancelist,
               'distance_charge_units' : dis_u,
               'nsteps' : self.ctx.points,
               'guess' : self.ctx.guess,
               'stepsize' : self.ctx.step,
               #'fitresults' : [a, latticeconstant, c],
               #'fit' : fit_new,
               'residuals' : residuals,
               'bulk_deriv' : bulk_deriv,
               'bulk_modulus' : bulk_modulus * 160.217733,#* echarge * 1.0e21,#GPa
               'bulk_modulus_units' : 'GPa',
               'successful' : self.ctx.successful,
               'info' : self.ctx.info,
               'warnings' : self.ctx.warnings,
               'errors' : self.ctx.errors}
               
        if self.ctx.successful:
            self.report('Done, Equation of states calculation complete')
        else:
            self.report('Done, but something went wrong.... Properly some individual calculation failed or a scf-cylcle did not reach the desired distance.')

        # output must be aiida Data types.
        outnode = ParameterData(dict=out)
        outnodedict['results_node'] = outnode

        # create links between all these nodes...
        outputnode_dict = create_eos_result_node(**outnodedict)
        outputnode = outputnode_dict.get('output_eos_wc_para')
        outputnode.label = 'output_eos_wc_para'
        outputnode.description = 'Contains equation of states results and information of an fleur_eos_wc run.'

        outputstructure = outputnode_dict.get('gs_structure')
        outputstructure.label = 'ouput_eos_wc_structure'
        outputstructure.description = 'Structure with the scaling/volume of the lowest total energy extracted from fleur_eos_wc'

        returndict = {}
        returndict['output_eos_wc_para'] = outputnode
        returndict['output_eos_wc_structure'] = outputstructure

        # create link to workchain node
        for link_name, node in returndict.iteritems():
            self.out(link_name, node)

            
    def control_end_wc(self, errormsg):
        """
        Controled way to shutdown the workchain. will initalize the output nodes
        The shutdown of the workchain will has to be done afterwards
        """
        self.ctx.successful = False
        self.ctx.abort = True
        self.report(errormsg) # because return_results still fails somewhen
        self.ctx.errors.append(errormsg)
        self.return_results()
        
        return

if __name__ == "__main__":
    import argparse

    # TODO: this is not usable in the command line, since you cannot provide the nodes in the shell
    # instead you need to use somehting that will load the nodes from pks or uuids
    # checkout what other aiida people do here
    parser = argparse.ArgumentParser(description='Equation of states workchain with Fleur. Does scf-cycles for a structure with different scaleings.')
    parser.add_argument('--wf_para', type=ParameterData, dest='wf_parameters',
                        help='Parameter data node, specifing workflow parameters', required=False)
    parser.add_argument('--inpgen', type=Code, dest='inpgen',
                        help='The inpgen code node to use', required=True)
    parser.add_argument('--fleur', type=Code, dest='fleur',
                        help='The FLEUR code node to use', required=True)
    parser.add_argument('--structure', type=StructureData, dest='structure',
                        help='The crystal structure node', required=True)
    parser.add_argument('--calc_para', type=ParameterData, dest='calc_parameters',
                        help='Parameters for the FLEUR calculation', required=False)
    args = parser.parse_args()
    res = fleur_eos_wc.run(wf_parameters=args.wf_parameters,
                           structure=args.structure,
                           calc_parameters=args.calc_parameters,
                           inpgen=args.inpgen, fleur=args.fleur)


@wf
def create_eos_result_node(**kwargs):
    """
    This is a pseudo wf, to create the rigth graph structure of AiiDA.
    This wokfunction will create the output node in the database.
    It also connects the output_node to all nodes the information commes from.
    So far it is just also parsed in as argument, because so far we are to lazy
    to put most of the code overworked from return_results in here.
    """
    outdict = {}
    outpara = kwargs.get('results_node', {})
    outdict['output_eos_wc_para'] = outpara.copy()#clone()
    # copy, because we rather produce the same node twice
    # then have a circle in the database for now...
    outputdict = outpara.get_dict()
    structure = load_node(outputdict.get('initial_structure'))
    gs_scaling = outputdict.get('scaling_gs', 0)
    if gs_scaling:
        gs_structure = rescale(structure, Float(gs_scaling))
        outdict['gs_structure'] = gs_structure

    return outdict


def eos_structures(inp_structure, scalelist):
    """
    Creates many rescalled StrucutureData nodes out of a crystal structure.
    Keeps the provanance in the database.

    :param StructureData, a StructureData node (pk, sor uuid)
    :param scalelist, list of floats, scaling factors for the cell

    :returns: list of New StructureData nodes with rescalled structure, which are linked to input Structure
    """
    structure = is_structure(inp_structure)
    if not structure:
        #TODO: log something (test if it gets here at all)
        return None
    re_structures = []

    for scale in scalelist:
        s = rescale(structure, Float(scale)) # this is a wf
        re_structures.append(s)

    return re_structures

'''
def fit_latticeconstant(scale, eT):
    """
    Extract the lattice constant out of an parabola fit.

    scale : list of scales, or lattice constants
    eT: list of total energies
    """
    # TODO Fit teh real function Mun... not a parabola
    import numpy as np
    # call fitt pol2 # or something else
    #def func(x, a, b, c):
    #    return a*x**2 + b*x + c
    f1 = np.polyfit(scale,eT,2)
    a0 = f1[0]
    a1 = f1[1]
    a2 = f1[2]
    la = -0.5*a1/a0
    c = a2 - a1**2/a2
    return a0,la,c, f1

def parabola(x, a, b, c):
    return a*x**2 + b*x + c
'''

# TODO other fits
def birch_murnaghan_fit(energies, volumes):
    """
    least squares fit of a Birch-Murnaghan equation of state curve. From delta project
    containing in its columns the volumes in A^3/atom and energies in eV/atom
    # The following code is based on the source code of eos.py from the Atomic
    # Simulation Environment (ASE) <https://wiki.fysik.dtu.dk/ase/>.
    :params energies: list (numpy arrays!) of total energies eV/atom
    :params volumes: list (numpy arrays!) of volumes in A^3/atom

    #volume, bulk_modulus, bulk_deriv, residuals = Birch_Murnaghan_fit(data)
    """
    fitdata = np.polyfit(volumes[:]**(-2./3.), energies[:], 3, full=True)
    ssr = fitdata[1]
    sst = np.sum((energies[:] - np.average(energies[:]))**2.)
    #print(fitdata)
    #print(ssr)
    #print(sst)
    residuals0 = ssr/sst
    deriv0 = np.poly1d(fitdata[0])
    deriv1 = np.polyder(deriv0, 1)
    deriv2 = np.polyder(deriv1, 1)
    deriv3 = np.polyder(deriv2, 1)

    volume0 = 0
    x = 0
    for x in np.roots(deriv1):
        if x > 0 and deriv2(x) > 0:
            volume0 = x**(-3./2.)
            break

    if volume0 == 0:
        print('Error: No minimum could be found')
        exit()

    derivV2 = 4./9. * x**5. * deriv2(x)
    derivV3 = (-20./9. * x**(13./2.) * deriv2(x) -
               8./27. * x**(15./2.) * deriv3(x))
    bulk_modulus0 = derivV2 / x**(3./2.)
    #print('bulk modulus 0: {} '.format(bulk_modulus0))
    bulk_deriv0 = -1 - x**(-3./2.) * derivV3 / derivV2

    return volume0, bulk_modulus0, bulk_deriv0, residuals0

def birch_murnaghan(volumes, volume0, bulk_modulus0, bulk_deriv0):
    """
    This evaluates the Birch Murnaghan equation of states
    """
    PV = []
    EV = []
    v0 = volume0
    bm = bulk_modulus0
    dbm = bulk_deriv0

    for vol in volumes:
        pv_val = 3 * bm/2. * ((v0/vol)**(7/3.) - (v0/vol)**(5/3.)) * (1 + 3/4. * (dbm -4) * ((v0/vol)**(2/3.)-1))
        PV.append(pv_val)
        ev_val = 9 * bm*v0/16. * ((dbm*(v0/vol)**(2/3.) - 1)**(3) * ((v0/vol)**(2/3.)-1)**2 * (6-4*(v0/vol)**(2/3.)))
        EV.append(ev_val)
    return EV, PV



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/broeder-j/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################

"""
In this module you find the workflow 'fleur_eos_wc_simple' for the calculation of
of an equation of state, with only the structure and optional wc parameters as input
"""

import numpy as np
from aiida.orm import Code, DataFactory, load_node
from aiida.orm.data.base import Float
#from aiida.work.process_registry import ProcessRegistry
from aiida.work.workchain import WorkChain, ToContext#,Outputs
#from aiida.work import workfunction as wf
from aiida.work.run import submit
#from aiida_fleur.tools.StructureData_util import rescale, is_structure
from aiida_fleur.workflows.scf import fleur_scf_wc
from aiida_fleur.workflows.optimize_para import fleur_optimize_parameters_wc
from aiida_fleur.workflows.eos import fleur_eos_wc, eos_structures
from aiida_fleur.tools.common_fleur_wf import test_and_get_codenode


StructureData = DataFactory('structure')
ParameterData = DataFactory('parameter')
FleurInpData = DataFactory('fleur.fleurinp')


class fleur_eos_wc_simple(WorkChain):
    """
    This workflow prepares the input parameters for an fleur_eos_wc and
    calls it for the calculation of the equation of states of a structure.


    :param wf_parameters: ParameterData node, optional, protocol specification will be parsed like this to fleur_eos_wc
    :param structure: StructureData node, bulk crystal structure
    :param inpgen: Code node,
    :param fleur: Code node,


    :return output_eos_wc_para: ParameterData node of fleur_eos_wc, contains relevant output information.
    about general succces, fit results and so on.

    """

    _workflowversion = "0.1.0"

    def __init__(self, *args, **kwargs):
        super(fleur_eos_wc_simple, self).__init__(*args, **kwargs)

    @classmethod
    def define(cls, spec):
        super(fleur_eos_wc_simple, cls).define(spec)
        spec.input("wf_parameters", valid_type=ParameterData, required=False,
                   default=ParameterData(dict={
                       'fleur_runmax': 4,
                       'points' : 9,
                       'step' : 0.002,
                       'guess' : 1.00,
                       'resources' : {"num_machines": 1},#, "num_mpiprocs_per_machine" : 12},
                       'walltime_sec':  60*60,
                       'queue_name' : '',
                       'custom_scheduler_commands' : ''}))
        spec.input("structure", valid_type=StructureData, required=True)
        spec.input("inpgen", valid_type=Code, required=True)
        spec.input("fleur", valid_type=Code, required=True)
        spec.outline(
            cls.start,
            cls.determine_parameters,
            cls.run_eos_wc,
            cls.inspect_eos_wc,# or maybe merge with last step?
            cls.return_results
        )
        #spec.dynamic_output()

    # first we initialize everything,
    # get structure with lowest volume
    # find out the optiomal parameter for these structures.
    # then make as modular as possible


    def start(self):
        """
        check parameters, what condictions? complete?
        check input nodes
        """
        self.report('started simple eos workflow version {}'.format(self._workflowversion))
        #self.report("Workchain node identifiers: {}".format(ProcessRegistry().current_calc_node))

        ### input check ###

        # initialize contexts

        self.ctx.last_calc2 = None
        self.ctx.calcs = []
        self.ctx.calcs_future = []
        self.ctx.structures = []
        self.ctx.temp_calc = None
        self.ctx.structurs_uuids = []
        self.ctx.scalelist = []
        self.ctx.volume = []
        self.ctx.volume_peratom = []
        self.ctx.org_volume = -1# avoid div 0
        self.ctx.labels = []
        self.ctx.successful = True


        # Check on inputnodes

        inputs = self.inputs


        # wf_parameters:

        wf_dict = inputs.wf_parameters.get_dict()

        # set values, or DEFAULTS
        self.ctx.points = wf_dict.get('points', 9)
        self.ctx.step = wf_dict.get('step', 0.002)
        self.ctx.guess = wf_dict.get('guess', 1.00)
        self.ctx.serial = wf_dict.get('serial', False)
        self.ctx.custom_scheduler_commands = wf_dict.get('custom_scheduler_commands', '')
        self.ctx.max_number_runs = wf_dict.get('fleur_runmax', 4)


        # codes
        if 'inpgen' in inputs:
            try:
                test_and_get_codenode(inputs.inpgen, 'fleur.inpgen', use_exceptions=True)
            except ValueError:
                error = ("The code you provided for inpgen of FLEUR does not "
                         "use the plugin fleur.inpgen")
                self.control_end_wc(error)
                self.abort(error)

        if 'fleur' in inputs:
            try:
                test_and_get_codenode(inputs.fleur, 'fleur.fleur', use_exceptions=True)
            except ValueError:
                error = ("The code you provided for FLEUR does not "
                         "use the plugin fleur.fleur")
                self.control_end_wc(error)
                self.abort(error)

    def determine_parameters(self):
        """
        determine the optimal input parameters for the given structure
        """

        points = self.ctx.points
        step = self.ctx.step
        guess = self.ctx.guess
        startscale = guess-(points-1)/2*step

        for point in range(points):
            self.ctx.scalelist.append(startscale + point*step)

        smallest_structure = eos_structures(self.inputs.structure, self.ctx.scalelist[0])[0]


        optimize_res = submit(fleur_optimize_parameters_wc,
                                    structure=smallest_structure,
                                    inpgen=self.inputs.inpgen,
                                    fleur=self.inputs.fleur)
        #wf_parameters

        return ToContext(optimize_res=optimize_res)

    def run_eos_wc(self):
        """
        Run the fleur_eos_wc
        """

        structure = self.inputs.structure
        eos_wc_para = self.inputs.wf_parameters
        optimize_res = self.ctx.optimize_res
        calc_parameters = optimize_res.get('optimized_para')
        fleur = self.inputs.fleur
        inpgen = self.inputs.inpgen

        form = structure.get_formula()
        label = 'fleur_eos_wc on {}'.format(form)
        description = 'Fleur eos of {}'.format(form)

        eos_res = submit(fleur_eos_wc, wf_parameters=eos_wc_para, fleur=fleur, inpgen=inpgen,
                structure=structure, calc_parameters=calc_parameters,
                _label=label, _description=description)

        return ToContext(**eos_res)


    def inspect_eos_wc(self):
        """
        Check the results of the fleur_eos_wc
        """
        # TODO: check on the returns of fleur_eos_wc
        pass



    def return_results(self):
        """
        Return the result node of fleur_eos_wc
        """

        returndict = {}

        if self.ctx.successful:
            self.report('Done, Simple equation of states calculation complete')

            returndict['output_eos_wc_para'] = self.ctx.eos_res['output_eos_wc_para']
            returndict['output_eos_wc_structure'] = self.ctx.eos_res['output_eos_wc_structure']
        else:
            self.report('Done, but something failed in fleur_eos_wc.')


        # create link to workchain node
        for link_name, node in returndict.iteritems():
            self.out(link_name, node)



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/broeder-j/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################

"""
In here we put all things (methods) that are common to workflows AND DO NOT
depend on AiiDA classes, therefore can be used without loading the dbenv.
Util that does depend on AiiDA classes should go somewhere else.
"""

def convert_formula_to_formula_unit(formula):
    """
    Converts a formula to the smalles chemical formula unit
    'Be4W2' -> 'Be2W'
    """
    from aiida_fleur.tools.common_fleur_wf_util import get_natoms_element
    from fractions import gcd

    # get formula dict
    # find greatest common divider of values
    # form formula unit string
    element_count_dict = get_natoms_element(formula)
    nelements = element_count_dict.values()
    g = nelements[0]
    for a2 in nelements:
        g = gcd(g,a2)

    formula_unit_string = ''
    for key, val in element_count_dict.iteritems():
        new_val = int(val/g)
        if new_val == 1:
            new_val = ''
        formula_unit_string = formula_unit_string + '{}{}'.format(key, new_val)

    return formula_unit_string
# test convert_formula_to_formula_unit('Be4W2')
# Be2W


def get_natoms_element(formula):
    """
    Converts 'Be24W2' to {'Be': 24, 'W' : 2}, also BeW to {'Be' : 1, 'W' : 1}
    """

    import re
    elem_count_dict = {}
    elements = re.findall('[A-Z][^A-Z]*', formula)
    #re.split('(\D+)', formula)

    for i, elm in enumerate(elements):
        elem_count = re.findall('\d+|\D+', elm)
        #print(elem_count)
        if len(elem_count) == 1:
            elem_count_dict[elem_count[0]] = 1
        else:
            elem_count_dict[elem_count[0]] = float(elem_count[1])


    return elem_count_dict

# test
#get_natoms_element('BeW')
#get_natoms_element('Be2W')

def ucell_to_atompr(ratio, formulas, element, error_ratio=[]):
    """
    Converts unit cell ratios into atom ratios.

    len(ratio) == len(formulas) (== len(error_ratio))
    ucell_to_atompr([10, 1, 7], ['Be12Ti', 'Be17Ti2', 'Be2'], element='Be', [0.1, 0.1, 0.1])
    """
    import numpy as np


    atompro = []
    atompro_err = []
    if not (len(ratio) == len(formulas)):
        return

    n_atoms_formula = []
    for formula in formulas:
        res = get_natoms_element(formula)
        n_atoms_formula.append(res.get(element, 0))

    atompro = np.array(ratio)*np.array(n_atoms_formula)
    total = sum(atompro)
    atompro = atompro/total

    if len(error_ratio):
        atompro_err_t = np.array(error_ratio)*np.array(n_atoms_formula)
        e_sum = np.sqrt(sum(atompro_err_t**2))
        atompro_err = 1/total*(np.sqrt(atompro_err_t**2 + (atompro*e_sum)**2))

    return atompro, atompro_err

# test
#print(ucell_to_atompr([10, 1, 7], ['Be12Ti', 'Be17Ti2', 'Be2'], element='Be'))
#print(ucell_to_atompr([10, 1, 7], ['Be12Ti', 'Be17Ti2', 'Be2'], element='Be', error_ratio=[0.1,0.1,0.1]))
#(array([ 0.79470199,  0.11258278,  0.09271523]), [])
#(array([ 0.79470199,  0.11258278,  0.09271523]), array([ 0.01357192,  0.01136565,  0.0018444 ]))


def calc_stoi(unitcellratios, formulas, error_ratio=[]):
    """
    Calculate the Stoichiometry with errors from a given unit cell ratio, formulas.

    Example:
    calc_stoi([10, 1, 7], ['Be12Ti', 'Be17Ti2', 'Be2'], [0.1, 0.01, 0.1])
    ({'Be': 12.583333333333334, 'Ti': 1.0}, {'Be': 0.12621369924887876, 'Ti': 0.0012256517540566825})
    calc_stoi([10, 1, 7], ['Be12Ti', 'Be17Ti2', 'Be2'])
    ({'Be': 12.583333333333334, 'Ti': 1.0}, {})
    """

    import numpy as np

    stoi = {}
    if not (len(unitcellratios) == len(formulas)):
        return

    errors_stoi = {}
    for i, formula in enumerate(formulas):
        res = get_natoms_element(formula)
        for element, val in res.iteritems():
            stoi_elm = stoi.get(element, 0)
            stoi[element] = stoi_elm + val*unitcellratios[i]
            if len(error_ratio):
                errors = errors_stoi.get(element, 0)
                errors_stoi[element] = errors + val*val*error_ratio[i]*error_ratio[i]

    # make smallest number always one.
    vals = stoi.values()
    minv = min(vals)
    keymin = stoi.keys()[vals.index(minv)]
    norm_stoi = {}
    for key, val in stoi.iteritems():
        norm_stoi[key] = stoi[key]/minv
        if len(error_ratio):
            errors_stoi[key] = 1/stoi[keymin]*np.sqrt((errors_stoi[key]**2 + (stoi[key]/stoi[keymin]*errors_stoi[keymin])**2))
    return norm_stoi, errors_stoi

# test
#print(calc_stoi([10, 1, 7], ['Be12Ti', 'Be17Ti2', 'Be2'], [0.1, 0.01, 0.1]))
#print(calc_stoi([10, 1, 7], ['Be12Ti', 'Be17Ti2', 'Be2']))
#({'Be': 12.583333333333334, 'Ti': 1.0}, {'Be': 0.12621369924887876, 'Ti': 0.0012256517540566825})
#({'Be': 12.583333333333334, 'Ti': 1.0}, {})


def get_atomprocent(formula):
    """
    This converts a formula to a dictionary with elemnt : atomprocent
    example converts 'Be24W2' to {'Be': 24/26, 'W' : 2/26}, also BeW to {'Be' : 0.5, 'W' : 0.5}
    :params: formula: string
    :returns: a dict, element : atomprocent

    # Todo alternative with structuredata
    """
    form_dict_new = {}
    form_dict = get_natoms_element(formula)
    ntotal = sum(form_dict.values())
    for key, val in form_dict.iteritems():
        val_new = float(val)/ntotal
        form_dict_new[key] = val_new
    return form_dict_new

# test




def get_weight_procent(formula):
    """
    This converts a formula to a dictionary with elemnt : weightprocent
    example converts 'Be24W2' to {'Be': , 'W' : }, also BeW to {'Be' : , 'W' : }
    :params: formula: string
    :returns: a dict, element : weightprocent

    # Todo alternative with structuredata
    """

    pass


#def norm_total_energy_peratom(totalenergy, formula)
#def norm_total_energy_perunitcell(totalenergy, )

def determine_formation_energy(struc_te_dict, ref_struc_te_dict):
    """
    This method determines the formation energy.
    E_form =  E(A_xB_y) - x*E(A) - y*E(B)

    :inputs: struc_te_dict: python dictionary in the form of {'formula' : total_energy} for the compound(s)
    :inputs: ref_struc_te_dict: python dictionary in the form of {'formula' : total_energy per atom, or per unit cell} for the elements
    (if the formula of the elements contains a number the total energy is devided by that number)
    :returns: list of floats, dict {formula : eform, ..} units energy/per atom, energies have some unit as energies given
    """
    eform_list = []
    eform_dict = {}
    #ref_el = ref_struc_te_dict.keys()
    ref_struc_te_dict_norm = ref_struc_te_dict#{}
    # assume reference to be normalized

    # normalize reference
    #for key, val in ref_struc_te_dict.iteritems():
    ##    elem_n = get_natoms_element(key)
    #    ref_struc_te_dict_norm[elem_n.keys()[0]] = val / elem_n.values()[0]
    ref_el_norm = ref_struc_te_dict_norm.keys()

    for formula, tE in struc_te_dict.iteritems():
        elements_count = get_natoms_element(formula)
        ntotal = float(sum(elements_count.values()))
        print(ntotal)
        eform = tE#abs(tE)
        for elem, count in elements_count.iteritems():
            if elem in ref_el_norm:
                eform = eform - count * ref_struc_te_dict_norm.get(elem)#abs(ref_struc_te_dict.get(elem))
            else:
                print('Reference energy missing for element {}. '
                      'You need to provide reference energies for all elements in you compound.'
                       ''.format(elem))
        eform_dict[formula] = eform/ntotal
        eform_list.append(eform/ntotal)
    return eform_list, eform_dict

# test
#determine_formation_energy({'BeW' : 2, 'Be2W' : 2.5}, {'Be' : 1, 'W' : 1})

def determine_convex_hull(formation_en_grid):
    """
    Wraps the scipy.spatial ConvexHull algo for our purposes.
    For now only for 2D phase diagrams
    Adds the points [1.0, 0.0] and [0.0, 1.0], because in material science these
    are always there.

    :params: formation_en_grid: list of points in phase space [[x, formation_energy]]
    :returns: a hul datatype
    """
    import numpy as np
    from scipy.spatial import ConvexHull

    # TODO multi d
    # check if endpoints are in
    if [1.0, 0.0] not in formation_en_grid:
        formation_en_grid.append([1.0, 0.0])
    if [0.0, 0.0] not in formation_en_grid:
        formation_en_grid.append([0.0, 0.0])

    points = np.array(formation_en_grid)
    hull = ConvexHull(points)

    return hull


def inpgen_dict_set_mesh(inpgendict, mesh):
    """
    params: python dict, used for inpgen parameterdata node
    params: mesh either as returned by kpointsdata or tuple of 3 integers

    returns: python dict, used for inpgen parameterdata node
    """
    if len(mesh) == 2:
        kmesh = mesh[0]
    elif len(mesh) == 3:
        kmesh = mesh
    kpt_dict = inpgendict.get('kpt', {})
    kpt_dict['div1'] = kmesh[0]
    kpt_dict['div2'] = kmesh[1]
    kpt_dict['div3'] = kmesh[2]

    inpgendict_new = inpgendict
    inpgendict_new['kpt'] = kpt_dict

    return inpgendict_new

# test
#inpgen_dict_set_mesh(Be_para.get_dict(), mesh)

def powerset(L):
    """
    Constructs the power set, 'potenz Menge' of a given list.

    return list: of all possible subsets
    """
    import itertools
    pset = []
    for n in xrange(len(L) + 1):
        for sset in itertools.combinations(L, n):
            pset.append(sset)
    return pset

#a = powerset([1, 2, 3, 4])
#a = powerset(['Be', 'Be2W', 'Be12W', 'Be22W', 'W'])
#print(a)
#print(len(a))

def determine_reactions(formula, available_data):
    """
    Determines and balances theoretical possible reaction.

    Stoichiometry
    'Be12W', [Be12W, Be2W, Be, W, Be22W] -> [[Be22W+Be2W], [Be12W], [Be12+W],...]
    params formula: string, given educts (left side of equation),
    params available_data: list of strings of compounds (products), from which all possibilities will be constructed

    """

    # 1. for each compound try to balance equation
    # 2. for each compound with any other compound in list, try to balance equation
    # 3. for each compound with each two other compounds ... till other compounds
    reactions = []
    constructed_products = []
    pset_available_data = powerset(available_data) # if len available_data to large cut off?
    for i,dataset in enumerate(pset_available_data):
        productstring = ''
        if len(dataset) <1:
            continue
        for entry in dataset:
            productstring = productstring + '{}+'.format(entry)

        productstring = productstring[:-1]
        constructed_products.append(productstring)
        pos_reaction = '{}->{}'.format(formula, productstring)
        bal_reaction = balance_equation(pos_reaction, allow_negativ=False, allow_zero=False, eval_linear=True)
        # We do not allow zero coefficents of products, because the resulting equation should already be in our list.
        if bal_reaction:
            # TODO post process (i.e are remove 0 compounds)
            reactions.append(bal_reaction)
        else:
            continue
    return reactions

# test reac = determine_reactions('Be12W', ['Be12W', 'Be2W', 'Be', 'W', 'Be22W'])
#print(reac ['1*Be12W->1*Be12W', '1*Be12W->1*Be2W+10*Be', '2*Be12W->1*Be2W+1*Be22W',
#             '1*Be12W->12*Be+1*W', '11*Be12W->5*W+6*Be22W'])

#reac = determine_reactions('Be12Ti', ['Be12Ti', 'Be17Ti2', 'BeTi', 'Ti', 'Be',
#                                       'Be2Ti', 'Be8Ti4'])
#print(reac ['1*Be12Ti->1*Be12Ti', '2*Be12Ti->1*Be17Ti2+7*Be', '1*Be12Ti->1*BeTi+11*Be',
#             '1*Be12Ti->1*Ti+12*Be', '1*Be12Ti->10*Be+1*Be2Ti', '4*Be12Ti->40*Be+1*Be8Ti4'])


def convert_eq_to_dict(equationstring):
    """
    Converts an equation string to a dictionary
    convert_eq_to_dict('1*Be12Ti->10*Be+1*Be2Ti+5*Be') ->
    {'products': {'Be': 15, 'Be2Ti': 1}, 'educts': {'Be12Ti': 1}}
    """
    eq_dict = {'products': {}, 'educts' : {}}
    product_dict = {}
    educt_dict = {}

    eq_split = equationstring.split('->')
    products = eq_split[1].split('+')
    educts = eq_split[0].split('+')

    for product in products:
        p_list = product.split('*')
        product_dict[p_list[-1]] = int(p_list[0]) + product_dict.get(p_list[-1], 0)
    for educt in educts:
        e_list = educt.split('*')
        educt_dict[e_list[-1]] = int(e_list[0]) + educt_dict.get(e_list[-1], 0)

    eq_dict['products'] = product_dict
    eq_dict['educts'] = educt_dict
    return eq_dict

# test convert_eq_to_dict('1*Be12Ti->10*Be+1*Be2Ti+5*Be')
 # {'products': {'Be': 15, 'Be2Ti': 1}, 'educts': {'Be12Ti': 1}}


def get_enhalpy_of_equation(reaction, formenergydict):
    """
    calculate the enhalpy per atom of a given reaction from the given data.

    param reaction: string
    param fromenergydict: dictionary that contains the {compound: formationenergy per atom}

    # TODO check if physics is right
    """
    reac_dict = convert_eq_to_dict(reaction)
    educt_energy = 0
    product_energy = 0

    for compound, factor in reac_dict.get('educts', {}).iteritems():
        compound_e = 0
        try:
            compound_e = formenergydict.get(compound)
        except KeyError:
            print('Formation energy of compound {} not given in {}.'
                  'I abort...'.format(compound, formenergydict))
            compound_e = 0
            return None
        educt_energy = educt_energy + factor*compound_e

    for compound, factor in reac_dict.get('products', {}).iteritems():
        try:
            compound_e = formenergydict.get(compound)
        except KeyError:
            print('Formation energy of compound {} not given in {}.'
                  'I abort...'.format(compound, formenergydict))
            compound_e = 0
            return None
        product_energy = product_energy + factor*compound_e


    return educt_energy - product_energy

from sympy import Symbol #somehow, otherwise it in not found in balance equation...

def balance_equation(equation_string, allow_negativ=False, allow_zero=False, eval_linear=True):
    """
    Method that balances a chemical equation.

    param equation_string: string (with '->')
    param allow_negativ: bool, default False, allows for negative coefficents for the products.

    return string: balanced equation

    balance_equation("C7H16+O2 -> CO2+H2O"))
    balance_equation("Be12W->Be22W+Be12W")
    balance_equation("Be12W->Be12W")

    1*C7H16+11*O2 ->7* CO2+8*H2O
    None
    1*Be12W->1*Be12W
    #TODO The solver better then what we need. Currently if system is over
    #"Be12W->Be2W+W+Be" solves to {a: 24, b: -d/2 + 144, c: d/2 - 120}-> FAIL-> None
    # The code fails in the later stage, but this solution should maybe be used.

    code adapted from stack exchange (the messy part):
    https://codegolf.stackexchange.com/questions/8728/balance-chemical-equations
    """

    import sys,re
    from sympy.solvers import solve
    from fractions import gcd
    from collections import defaultdict
    letters = 'abcdefghijklmnopqrstuvwxyz'
    Ls=list(letters)
    eq=equation_string
    Ss,Os,Es,a,i = defaultdict(list),Ls[:],[],1,1
    for p in eq.split('->'):
        for k in p.split('+'):
            c = [Ls.pop(0), 1]
            for e,m in re.findall('([A-Z][a-z]?)([0-9]*)',k):
                m=1 if m=='' else int(m)
                a*=m
                d=[c[0],c[1]*m*i]
                Ss[e][:0],Es[:0]=[d],[[e,d]]
        i=-1
    Ys=dict((s,eval('Symbol("'+s+'")')) for s in Os if s not in Ls)
    Qs=[eval('+'.join('%d*%s'%(c[1],c[0]) for c in Ss[s]),{},Ys) for s in Ss]+[Ys['a']-a]
    k=solve(Qs,*Ys)
    if k:# if a solution is found multiply by gcd
        # TODO? check if solution has linear dependence: and evaluate
        #for c in k.values():
        #    for char in list(letters):
        #        if char in str(c):
        #             pass
        N=[]#[k[Ys[s]]for s in sorted(Ys)]
        for s in sorted(Ys):
            n = k[Ys[s]]
            # idea: check if char in n, then linear depended, then
            try: # since solver gives also a linear depended solution if correct, but code fails then
                if n<0 and not allow_negativ: # We allow for 0 but might be an other case to think about
                    return None
            except TypeError:
                return None # TODO Maybe other return value... maybe list of some values for
                # linear dependencies, d,e,....? also choose them that the value is positive...?
            if n==0 and not allow_zero:
                return None
            N.append(n)
        g = N[0]
        for a1, a2 in zip(N[0::2],N[0::2]):
            g=gcd(g,a2)
        N = [i/g for i in N]
        pM = lambda c: str(c)+'*'# if c!=1 else ''
        return '->'.join('+'.join(pM(N.pop(0))+str(t) for t in p.split('+')) for p in eq.split('->'))
    else:
        return None


# test
#print(balance_equation("C7H16+O2 -> CO2+H2O"))
#print(balance_equation("Be12W->Be2W+W+Be")#+Be12W+Be+Be22W"))
#print(balance_equation("Be12WO->Be2WO+W+Be+O2")#+Be12W+Be+Be22W"))
#print(balance_equation("Be12W->Be22W+Be12W"))
#print(balance_equation("Be12W->Be12W"))

#1*C7H16+11*O2 ->7* CO2+8*H2O
#None
#None
#None
#1*Be12W->1*Be12W

def check_eos_energies(energylist):
    """
    Checks if there is an abnormality in the total energies from the Equation of states.
    i.e. if one point has a larger energy then its two neighbors

    :param energylist: list of floats
    :returs nnormalies: integer
    """
    # TODO: possible improvements look at all differences, and introduce a threshold
    # Also look at several points...
    abnormalityindexlist = []
    abnormality = False
    i = 0
    for x, y, z in zip(energylist, energylist[1:], energylist[2:]):
        i = i + 1
        if x < y > z:
            abnormality = True
            abnormalityindexlist.append(i)
            print(x,y,z)
            print('annormly detected')

    return abnormality, abnormalityindexlist

#total_energy = [ -1, -2, -3 ,-2,-4,-3,-2,-1]
#check_eos_energies(total_energy)
#(True, [3])
#total_energy = [ -1, -2, -3 ,-2,-2,-3,-2,-1]
#check_eos_energies(total_energy)
#(False, [])
#total_energy = [ -1, -2, -3 ,-4,-5,-3,-2,-1]
#(False, [])




# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/broeder-j/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################

"""
This module, contains the crystal structure relaxation workflow for FLEUR.
"""

#TODO: print more user info
#  allow different inputs, make things optional(don't know yet how)
#  half number of iteration if you are close to be converged. (therefore one can start with 18 iterations, and if thats not enough run agian 9 or something)

#import sys,os
#from ase import *
#from ase.lattice.surface import *
#from ase.io import *
from aiida.orm import Code, DataFactory, load_node
from aiida.work.workchain import WorkChain, while_, if_, ToContext
from aiida.work.run import run, submit
from aiida_fleur.calculation.fleurinputgen import FleurinputgenCalculation
from aiida_fleur.calculation.fleur import FleurCalculation
from aiida_fleur.workflows.scf import fleur_scf_wc


StructureData = DataFactory('structure')
ParameterData = DataFactory('parameter')
FleurInpData = DataFactory('fleur.fleurinp')

class fleur_relax_wc(WorkChain):
    """
    This workflow relaxes a structure with Fleur calculation.

    :Params: a parameterData node,
    :returns: Success, last result node, list with convergence behavior
    """
    # wf_parameters: { wf_convergence_para:{  'density_criterion', 'energy_criterion'} #'converge_density' = True, 'converge_energy'= True}, max_force_cycle, force_criterion
    # calc_parameters: Parameter data for inpgen calculation
    @classmethod
    def _define(cls, spec):
        super(fleur_relax_wc, cls).define(spec)
        spec.input("wf_parameters", valid_type=ParameterData, required=True)
                   #, required=False, default = ParameterData(
                   #dict={'relax_runmax' : 4, 'fleur_runmax' : 10, 'density_criterion' : 0.00002 , 'converge_density': True, 'converge_energy' : True, 'energy_criterion' : 0.00002})# htr
                   #{default=make_str('quantumespresso.pw')))
        spec.input("structure", valid_type=StructureData)#, required=False
        spec.input("calc_parameters", valid_type=ParameterData)#, required=False
        #spec.input("parent_calculation", valid_type=JobCalculation)#, required=False # either fleur or inpgen.
        spec.input("inpgen", valid_type=Code)#, required=False
        #spec.input("computer", valid_type=Computer)#, required=False, default=get_computer_from_string('iff003')
        spec.input("fleur", valid_type=Code)#, required=True
        spec.outline(
            cls.start,
            cls.converge_scf,
            cls.calculate_forces,
            while_(cls.condition)(
                cls.converge_scf,
                cls.calculate_forces
                ),
            cls.return_results,
        )
        spec.dynamic_output()

    def start(self):
        """
        check parameters, what condictions? complete?
        check input nodes
        """
        ### input check ### ? or done automaticly, how optional?
        self.ctx.last_calc2 = None
        self.ctx.loop_count2 = 0
        self.ctx.calcs = []
        self.ctx.successful2 = False

        # set values, or defaults, default: always converge charge density, crit < 0.00002, max 4 fleur runs
        self.ctx.max_number_runs = self.inputs.wf_parameters.get_dict().get('relax_runmax', 4)
        self.ctx.max_force_cycle = self.inputs.wf_parameters.get_dict().get('max_force_cycle', 4)
        self.report('starting relaxation workchain')

    def converge_scf(self):
        """
        start scf-cycle from Fleur calculation
        """
        # run a convergence worklfow
        # Comment: Here we wait, because of run,
        # maybe using a future and submit instead might be better
        inputs = self.get_inputs_scf(self.ctx)

        if self.ctx.last_calc2:
            res = run(fleur_scf_wc,
                      wf_parameters=inputs['wf_parameters'],
                      fleurinp=inputs['fleurinp'],
                      fleur=inputs['fleur'])
                      #inputs)#
                  #wf_parameters=wf_para,
                  #structure=s,
                  #calc_parameters=parameters,
                  #inpgen = code,
                  #fleur=code2)#, computer=computer)
        else:
            res = run(fleur_scf_wc, wf_parameters=inputs['wf_parameters'],
                      structure=inputs['structure'], calc_parameters=inputs['calc_parameters'], inpgen = inputs['inpgen'], fleur=inputs['fleur'] )#inputs)#

        self.ctx.last_calc2 = res#.get('remote_folder', None)
        #return ResultToContext(last_calc2=res)

    def get_inputs_scf(self):
        """
        get the inputs for a scf-cycle
        """
        inputs = {}
        # produce the inputs for a convergence worklfow
        # if last calculation
        # create input from that
        if self.ctx.last_calc2:
            # get fleurinpData from inp_new.xml
            inputs['wf_parameters'] = self.inputs.wf_parameters
            inputs['fleur'] = self.inputs.fleur
            inputs['fleurinp'] = self.ctx['last_calc2']['fleurinpData']
            #inputs.wf_parameters = self.inputs.wf_parameters#.get_dict()
            #inputs.fleur = self.inputs.fleur
            #inputs.fleurinp = self.ctx['last_calc22']['fleurinp']
            #inputs.parent_folder = self.ctx['last_calc2']['remote_folder']
        # if not use input given from workflow
        else:
            inputs['wf_parameters'] = self.inputs.wf_parameters
            inputs['structure'] = self.inputs.structure
            inputs['calc_parameters'] = self.inputs.calc_parameters
            inputs['inpgen'] = self.inputs.inpgen
            inputs['fleur'] = self.inputs.fleur
            #inputs.wf_parameters = self.inputs.wf_parameters#.get_dict()
            #inputs.structure = self.inputs.structure
            #inputs.calc_parameters = self.input.calc_parameters
            #inputs.inpgen = self.inputs.inpgen
            #inputs.fleur = self.inputs.fleur


        return inputs

    def calculate_forces(self):
        """
        starts a Fleur calculation which calculates forces.
        """
        # get converged calculation
        # create new calculation with l_f =T, gff allow for relaxation of certain
        # atomtyps or species
        # dont copy broyden files, copy cdn1?
        # run fleur
        self.ctx.loop_count2 = self.ctx.loop_count2 + 1
        last_calc2 = self.ctx.last_calc2
        # be careful, test if convergence success or not...
        fleurinp = last_calc2.get('fleurinp', None)
        if fleurinp:
            fleurinp_new = fleurinp.copy()
        else: # warning
            fleurinp_new = None
            self.report('no fleurinp data was found in last_calc2')
        if False: # TODO something other specified in wf parameters
            change_dict = {'l_f' : True}
        else: # relax every atom in all direction specified in inp.xml
            change_dict = {'l_f' : True} # for calculation of forces

        fleurinp_new.set_inpchanges(change_dict)
        #fleurinp_new.store()# needed?

        remote = last_calc2.get('remote_folder', None)

        # run fleur
        FleurProcess = FleurCalculation.process()
        inputs = FleurCalculation.process().get_inputs_template()

        #inputs.parent_folder = remote
        inputs.code = self.inputs.fleur
        inputs.fleurinpdata = fleurinp_new
        inputs.parent_folder = remote # we need to copy cnd1
        inputs._options.resources = {"num_machines": 1}
        inputs._options.max_wallclock_seconds = 30 * 60
        # if code local use
        #if self.inputs.fleur.is_local():
        #    inputs._options.computer = computer
        #else:
        #    inputs._options.queue_name = 'th1'
        inputs._options.withmpi = False # for now
        self.report('Relax structure with Fleur, cycle: {}'.format(self.ctx.loop_count2))
        future = self.submit(FleurProcess, inputs)

        self.ctx.calcs.append(future)

        return ToContext(last_calc2=future)

    def condition(self):
        """
        check convergence condition
        """
        #pass max_force_cycle, force_criterion
        forces_converged = False
        energy_converged = False
        largest_force = self.ctx.last_calc2['output_parameters'].get_dict().get('force_largest')
        self.report('largest_force: {}'.format(largest_force))
        if self.inputs.wf_parameters.get_dict().get('converge_density', True):
            if self.inputs.wf_parameters.get_dict().get('force_criterion', 0.00002) >= largest_force:
                forces_converged = True


        # TODO?
        if self.inputs.wf_parameters.get_dict().get('converge_energy', True):
            #if self.inputs.wf_parameters.get_dict().get('energy_criterion', 0.0002) >= self.ctx.last_calc2['output_parameters'].dict.energy_hartree:
                #last_energy: # energy converged
            energy_converged = True
        else:
            energy_converged = False

        if forces_converged and energy_converged:
            self.ctx.successful2 = True
            return False
        elif self.ctx.loop_count2 >= self.ctx.max_force_cycle:
            return False
        else:
            return True


    def return_results(self):
        """
        return the results of the calculations
        """
        largest_force = self.ctx.last_calc2['output_parameters'].get_dict().get('force_largest')

        if self.ctx.successful2:
            self.report('Done, Des wors, forces converged')
            self.report('Fleur converged the total forces after {} scf-force cycles to {} ""'.format(self.ctx.loop_count2, largest_force))
        else: # loopcount reached
            self.report('Done, I reached the number of forces cycles, system is not relaxed enough.')
            self.report('Fleur converged the total forces after {} scf-force cycles to {} ""'.format(self.ctx.loop_count2, largest_force))
        outdict = self.ctx.last_calc2
        for k, v in outdict.iteritems():
            self.out(k, v)        # return success, and the last calculation outputs
        # ouput must be aiida Data types.



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/broeder-j/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################

import pytest
import os

## Collect the input files
file_path1 = '../files/inpxml/'
file_path2 = '../files/nonvalid_inpxml'


inpxmlfilefolder = os.path.dirname(os.path.abspath(__file__))
inpxmlfilefolder_valid =  os.path.abspath(os.path.join(inpxmlfilefolder, file_path1))
inpxmlfilefolder_non_valid = os.path.abspath(os.path.join(inpxmlfilefolder, file_path2))

inpxmlfilelist = []
for subdir, dirs, files in os.walk(inpxmlfilefolder_valid):
    for file in files:
        if file.endswith('.xml'):
            inpxmlfilelist.append(os.path.join(subdir, file))

inpxmlfilelist2 = []
for subdir, dirs, files in os.walk(inpxmlfilefolder_non_valid):
    for file in files:
        if file.endswith('.xml'):
            inpxmlfilelist2.append(os.path.join(subdir, file))
##

### Tests

# Testing of the Fleurinp data type


# Test if valid fleurinp datas can be created for all inp.xml files in a given test folder
        
@pytest.mark.parametrize("inpxmlfilepath", inpxmlfilelist)
@pytest.mark.usefixtures("aiida_env")
def test_fleurinp_valid_inpxml(inpxmlfilepath, aiida_env):
    """
    test if inp.xml file is reconnized as valid by fleur
    """
    from aiida.orm import DataFactory
    
    fleurinp = DataFactory('fleur.fleurinp')
    fleurinp_tmp = fleurinp(files=[inpxmlfilepath])
    
    assert fleurinp_tmp._has_schema == True
    assert fleurinp._schema_file_path is not None
    assert fleurinp_tmp.inp_dict != {}
    fleurinp_tmp._validate # will throw validation error if not validating

    
@pytest.mark.parametrize("inpxmlfilepath", inpxmlfilelist2)
@pytest.mark.usefixtures("aiida_env")
def test_fleurinp_non_valid_inpxml(inpxmlfilepath):
    """
    test what happens if inp.xml file is not valid
    either an inputvalidationError should be thrown by fleurinpData or an 
    XMLSyntaxError error if the inp.xml does not correspond to the xml schema.
    """
    from aiida.common.exceptions import InputValidationError
    from lxml.etree import XMLSyntaxError
    from aiida.orm import DataFactory
    
    fleurinp = DataFactory('fleur.fleurinp')
    
    with pytest.raises((InputValidationError, XMLSyntaxError)) as error:
        fleurinp_tmp = fleurinp(files=[inpxmlfilepath])

        
# test kpoints and structure and parameter data extraction

@pytest.mark.parametrize("inpxmlfilepath", inpxmlfilelist)
@pytest.mark.usefixtures("aiida_env")
def test_fleurinp_kpointsdata_extraction(inpxmlfilepath):
    """
    Extract a kpointsData from the fleurinp data, i.e inp.xml and check if 
    the resulting node is a valid kpoints data
    """
    from aiida.orm import DataFactory
    
    KpointsData = DataFactory('array.kpoints')
    fleurinp = DataFactory('fleur.fleurinp')

    
    fleurinp_tmp = fleurinp(files=[inpxmlfilepath])
    kptsd = fleurinp_tmp.get_kpointsdata_nwf(fleurinp_tmp)
    
    if kptsd is not None:
        assert isinstance(kptsd, KpointsData)
    else:
        pass
        # What todo here, may test inpxml are with latnam definded, which does not work here.
        # or without a kpoint list. Therefore this test might let two much through


@pytest.mark.parametrize("inpxmlfilepath", inpxmlfilelist)
@pytest.mark.usefixtures("aiida_env")
def test_fleurinp_parameterdata_extraction(inpxmlfilepath):
    """
    Extract a ParameterData from the fleurinp data, i.e inp.xml and check if 
    the resulting node is a valid ParameterData. ggf if it can be used by inpgen  
    """
    from aiida.orm import DataFactory
    
    ParameterData = DataFactory('parameter')
    fleurinp = DataFactory('fleur.fleurinp')
    
    fleurinp_tmp = fleurinp(files=[inpxmlfilepath])
    param = fleurinp_tmp.get_parameterdata_nwf()
    
    assert isinstance(param, ParameterData)
    
    # ToDo check if it is also right for inpgen...


@pytest.mark.parametrize("inpxmlfilepath", inpxmlfilelist)
@pytest.mark.usefixtures("aiida_env")
def test_fleurinp_structuredata_extraction(inpxmlfilepath):
    """
    Extract a ParameterData from the fleurinp data, i.e inp.xml and check if 
    the resulting node is a valid ParameterData.
    """
    from aiida.orm import DataFactory
    
    StructureData = DataFactory('structure')
    fleurinp = DataFactory('fleur.fleurinp')
    
    fleurinp_tmp = fleurinp(files=[inpxmlfilepath])
    struc = fleurinp_tmp.get_structuredata_nwf()

    if struc is not None:
        assert isinstance(struc, StructureData)
    else:
        pass
        # What todo here, may test inpxml are with latnam definded, 
        # which does not work here.
        # But if something else fails also None return. T
        # Therefore this test might let two much through    

       
# Input Modification tests
@pytest.mark.parametrize("inpxmlfilepath", inpxmlfilelist)
@pytest.mark.usefixtures("aiida_env")
def test_fleurinp_single_value_modification(inpxmlfilepath):
    """
    set kmax, itmax, minDistance in inp.xml input file of fleurinpdata to 
    10.2, 99, 0.000001, then check if it everything set
    """
    from aiida.orm import DataFactory
    from aiida_fleur.data.fleurinpmodifier import FleurinpModifier

    
    fleurinp = DataFactory('fleur.fleurinp')
    
    fleurinp_tmp = fleurinp(files=[inpxmlfilepath])
    fleurinpmode = FleurinpModifier(fleurinp_tmp)
    # caps matters here!
    fleurinpmode.set_inpchanges({'itmax': 99, 'minDistance' : 0.01, 'Kmax' : 10.2})
    
    
    fleurinpmode.show(display=False, validate=True)
    out = fleurinpmode.freeze()
    
    assert isinstance(out, fleurinp)
    
    # TODO check if set right
    
@pytest.mark.parametrize("inpxmlfilepath", inpxmlfilelist)
@pytest.mark.usefixtures("aiida_env")
def test_fleurinp_first_species_modification(inpxmlfilepath):
    """
    Decrease the rmt of the first species by 10%, check if rmt was set
    """
    pass

